import json
import logging
import os
import warnings
from concurrent.futures import ThreadPoolExecutor, as_completed
from itertools import islice
from string import Template
from typing import Any, Dict, List, Optional, Tuple, Union

from lbox.exceptions import (
    InvalidQueryError,
    LabelboxError,
    ResourceCreationError,
    ResourceNotFoundError,
)  # type: ignore

import labelbox.schema.internal.data_row_uploader as data_row_uploader
from labelbox.orm import query
from labelbox.orm.comparison import Comparison
from labelbox.orm.db_object import DbObject, Deletable, Updateable
from labelbox.orm.model import Entity, Field, Relationship
from labelbox.pagination import PaginatedCollection
from labelbox.schema.data_row import DataRow
from labelbox.schema.export_filters import DatasetExportFilters, build_filters
from labelbox.schema.export_params import (
    CatalogExportParams,
    validate_catalog_export_params,
)
from labelbox.schema.export_task import ExportTask
from labelbox.schema.iam_integration import IAMIntegration
from labelbox.schema.identifiable import GlobalKey, UniqueId
from labelbox.schema.internal.data_row_upsert_item import (
    DataRowCreateItem,
    DataRowItemBase,
    DataRowUpsertItem,
)
from labelbox.schema.internal.datarow_upload_constants import (
    FILE_UPLOAD_THREAD_COUNT,
    UPSERT_CHUNK_SIZE_BYTES,
)
from labelbox.schema.task import DataUpsertTask, Task

logger = logging.getLogger(__name__)


class Dataset(DbObject, Updateable, Deletable):
    """A Dataset is a collection of DataRows.

    Attributes:
        name (str)
        description (str)
        updated_at (datetime)
        created_at (datetime)
        row_count (int): The number of rows in the dataset. Fetch the dataset again to update since this is cached.

        created_by (Relationship): `ToOne` relationship to User
        organization (Relationship): `ToOne` relationship to Organization
    """

    name = Field.String("name")
    description = Field.String("description")
    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    row_count = Field.Int("row_count")

    # Relationships
    created_by = Relationship.ToOne("User", False, "created_by")
    organization = Relationship.ToOne("Organization", False)
    iam_integration = Relationship.ToOne(
        "IAMIntegration", False, "iam_integration", "signer"
    )

    def data_rows(
        self,
        from_cursor: Optional[str] = None,
        where: Optional[Comparison] = None,
    ) -> PaginatedCollection:
        """
        Custom method to paginate data_rows via cursor.

        Args:
            from_cursor (str): Cursor (data row id) to start from, if none, will start from the beginning
            where (dict(str,str)): Filter to apply to data rows. Where value is a data row column name and key is the value to filter on.
            example: {'external_id': 'my_external_id'} to get a data row with external_id = 'my_external_id'


        NOTE:
            Order of retrieval is newest data row first.
            Deleted data rows are not retrieved.
            Failed data rows are not retrieved.
            Data rows in progress *maybe* retrieved.
        """

        page_size = 500  # hardcode to avoid overloading the server
        where_param = (
            query.where_as_dict(Entity.DataRow, where)
            if where is not None
            else None
        )

        template = Template(
            """query DatasetDataRowsPyApi($$id: ID!, $$from: ID, $$first: Int, $$where: DatasetDataRowWhereInput)  {
                        datasetDataRows(id: $$id, from: $$from, first: $$first, where: $$where)
                            {
                                nodes { $datarow_selections }
                                pageInfo { hasNextPage startCursor }
                            }
                        }
                    """
        )
        query_str = template.substitute(
            datarow_selections=query.results_query_part(Entity.DataRow)
        )

        params = {
            "id": self.uid,
            "from": from_cursor,
            "first": page_size,
            "where": where_param,
        }

        return PaginatedCollection(
            client=self.client,
            query=query_str,
            params=params,
            dereferencing=["datasetDataRows", "nodes"],
            obj_class=Entity.DataRow,
            cursor_path=["datasetDataRows", "pageInfo", "startCursor"],
        )

    def create_data_row(self, items=None, **kwargs) -> "DataRow":
        """Creates a single DataRow belonging to this dataset.
        >>> dataset.create_data_row(row_data="http://my_site.com/photos/img_01.jpg")

        Args:
            items: Dictionary containing new `DataRow` data. At a minimum,
                must contain `row_data` or `DataRow.row_data`.
            **kwargs: Key-value arguments containing new `DataRow` data. At a minimum,
                must contain `row_data`.

        Raises:
            InvalidQueryError: If both dictionary and `kwargs` are provided as inputs
            InvalidQueryError: If `DataRow.row_data` field value is not provided
                in `kwargs`.
            InvalidAttributeError: in case the DB object type does not contain
                any of the field names given in `kwargs`.
            ResourceCreationError: If data row creation failed on the server side.
        """
        invalid_argument_error = "Argument to create_data_row() must be either a dictionary, or kwargs containing `row_data` at minimum"

        if items is not None and len(kwargs) > 0:
            raise InvalidQueryError(invalid_argument_error)

        args = items if items is not None else kwargs

        file_upload_thread_count = 1
        completed_task = self._create_data_rows_sync(
            [args], file_upload_thread_count=file_upload_thread_count
        )

        res = completed_task.result
        if res is None or len(res) == 0:
            raise ResourceCreationError(
                f"Data row upload did not complete, task status {completed_task.status} task id {completed_task.uid}"
            )

        return self.client.get_data_row(res[0]["id"])

    def _create_data_rows_sync(
        self, items, file_upload_thread_count=FILE_UPLOAD_THREAD_COUNT
    ) -> "DataUpsertTask":
        if file_upload_thread_count < 1:
            raise ValueError(
                "file_upload_thread_count must be a positive integer"
            )

        task: DataUpsertTask = self.create_data_rows(
            items, file_upload_thread_count
        )
        task.wait_till_done()

        if task.has_errors():
            raise ResourceCreationError(
                f"Data row upload errors: {task.errors}", cause=task.uid
            )
        if task.status != "COMPLETE":
            raise ResourceCreationError(
                f"Data row upload did not complete, task status {task.status} task id {task.uid}"
            )

        return task

    def create_data_rows(
        self, items, file_upload_thread_count=FILE_UPLOAD_THREAD_COUNT
    ) -> "DataUpsertTask":
        """Asynchronously bulk upload data rows

        Args:
            items (iterable of (dict or str))

        Returns:
            Task representing the data import on the server side. The Task
            can be used for inspecting task progress and waiting until it's done.

        Raises:
            InvalidQueryError: If the `items` parameter does not conform to
                the specification above or if the server did not accept the
                DataRow creation request (unknown reason).
            ResourceNotFoundError: If unable to retrieve the Task for the
                import process. This could imply that the import failed.
            InvalidAttributeError: If there are fields in `items` not valid for
                a DataRow.
            ValueError: When the upload parameters are invalid

        NOTE  dicts and strings items can not be mixed in the same call. It is a responsibility of the caller to ensure that all items are of the same type.
        """

        if file_upload_thread_count < 1:
            raise ValueError(
                "file_upload_thread_count must be a positive integer"
            )

        # Usage example
        upload_items = self._separate_and_process_items(items)
        specs = DataRowCreateItem.build(self.uid, upload_items)
        return self._exec_upsert_data_rows(specs, file_upload_thread_count)

    def _separate_and_process_items(self, items):
        string_items = [item for item in items if isinstance(item, str)]
        dict_items = [item for item in items if isinstance(item, dict)]
        dict_string_items = []
        if len(string_items) > 0:
            dict_string_items = self._build_from_local_paths(string_items)
        return dict_items + dict_string_items

    def _build_from_local_paths(
        self,
        items: List[str],
        file_upload_thread_count=FILE_UPLOAD_THREAD_COUNT,
    ) -> List[dict]:
        uploaded_items = []

        def upload_file(item):
            item_url = self.client.upload_file(item)
            return {"row_data": item_url, "external_id": item}

        with ThreadPoolExecutor(file_upload_thread_count) as executor:
            futures = [
                executor.submit(upload_file, item)
                for item in items
                if isinstance(item, str) and os.path.exists(item)
            ]
            more_items = [future.result() for future in as_completed(futures)]
            uploaded_items.extend(more_items)

        return uploaded_items

    def data_rows_for_external_id(
        self, external_id, limit=10
    ) -> List["DataRow"]:
        """Convenience method for getting a multiple `DataRow` belonging to this
        `Dataset` that has the given `external_id`.

        Args:
            external_id (str): External ID of the sought `DataRow`.
            limit (int): The maximum number of data rows to return for the given external_id

        Returns:
            A list of `DataRow` with the given ID.

        Raises:
         lbox.exceptions.ResourceNotFoundError: If there is no `DataRow`
                in this `DataSet` with the given external ID, or if there are
                multiple `DataRows` for it.
        """
        DataRow = Entity.DataRow
        where = DataRow.external_id == external_id

        data_rows = self.data_rows(where=where)
        # Get at most `limit` data_rows.
        at_most_data_rows = list(islice(data_rows, limit))

        if not len(at_most_data_rows):
            raise ResourceNotFoundError(DataRow, where)
        return at_most_data_rows

    def data_row_for_external_id(self, external_id) -> "DataRow":
        """Convenience method for getting a single `DataRow` belonging to this
        `Dataset` that has the given `external_id`.

        Args:
            external_id (str): External ID of the sought `DataRow`.

        Returns:
            A single `DataRow` with the given ID.

        Raises:
            lbox.exceptions.ResourceNotFoundError: If there is no `DataRow`
                in this `DataSet` with the given external ID, or if there are
                multiple `DataRows` for it.
        """
        data_rows = self.data_rows_for_external_id(
            external_id=external_id, limit=2
        )
        if len(data_rows) > 1:
            logger.warning(
                "More than one data_row has the provided external_id : `%s`. Use function data_rows_for_external_id to fetch all",
                external_id,
            )
        return data_rows[0]

    def export(
        self,
        task_name: Optional[str] = None,
        filters: Optional[DatasetExportFilters] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> ExportTask:
        """
        Creates a dataset export task with the given params and returns the task.

        >>>     dataset = client.get_dataset(DATASET_ID)
        >>>     task = dataset.export(
        >>>         filters={
        >>>             "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "data_row_ids": [DATA_ROW_ID_1, DATA_ROW_ID_2, ...] # or global_keys: [DATA_ROW_GLOBAL_KEY_1, DATA_ROW_GLOBAL_KEY_2, ...]
        >>>         },
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     task.wait_till_done()
        >>>     task.result
        """
        task, _ = self._export(task_name, filters, params, streamable=True)
        return ExportTask(task)

    def export_v2(
        self,
        task_name: Optional[str] = None,
        filters: Optional[DatasetExportFilters] = None,
        params: Optional[CatalogExportParams] = None,
    ) -> Union[Task, ExportTask]:
        """
        Creates a dataset export task with the given params and returns the task.

        >>>     dataset = client.get_dataset(DATASET_ID)
        >>>     task = dataset.export_v2(
        >>>         filters={
        >>>             "last_activity_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "label_created_at": ["2000-01-01 00:00:00", "2050-01-01 00:00:00"],
        >>>             "data_row_ids": [DATA_ROW_ID_1, DATA_ROW_ID_2, ...] # or global_keys: [DATA_ROW_GLOBAL_KEY_1, DATA_ROW_GLOBAL_KEY_2, ...]
        >>>         },
        >>>         params={
        >>>             "performance_details": False,
        >>>             "label_details": True
        >>>         })
        >>>     task.wait_till_done()
        >>>     task.result
        """

        warnings.warn(
            "The method export_v2 for Dataset is deprecated and will be removed in the next major release. Use the export method instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        task, is_streamable = self._export(task_name, filters, params)
        if is_streamable:
            return ExportTask(task, True)
        return task

    def _export(
        self,
        task_name: Optional[str] = None,
        filters: Optional[DatasetExportFilters] = None,
        params: Optional[CatalogExportParams] = None,
        streamable: bool = False,
    ) -> Tuple[Task, bool]:
        _params = params or CatalogExportParams(
            {
                "attachments": False,
                "embeddings": False,
                "metadata_fields": False,
                "data_row_details": False,
                "project_details": False,
                "performance_details": False,
                "label_details": False,
                "media_type_override": None,
                "model_run_ids": None,
                "project_ids": None,
                "interpolated_frames": False,
                "all_projects": False,
                "all_model_runs": False,
                "predictions": False,
            }
        )
        validate_catalog_export_params(_params)

        _filters = filters or DatasetExportFilters(
            {
                "last_activity_at": None,
                "label_created_at": None,
                "data_row_ids": None,
                "global_keys": None,
            }
        )

        mutation_name = "exportDataRowsInCatalog"
        create_task_query_str = (
            f"mutation {mutation_name}PyApi"
            f"($input: ExportDataRowsInCatalogInput!)"
            f"{{{mutation_name}(input: $input){{taskId isStreamable}}}}"
        )
        media_type_override = _params.get("media_type_override", None)

        if task_name is None:
            task_name = f"Export v2: dataset - {self.name}"
        query_params: Dict[str, Any] = {
            "input": {
                "taskName": task_name,
                "filters": {
                    "searchQuery": {
                        "scope": None,
                        "query": None,
                    }
                },
                "isStreamableReady": True,
                "params": {
                    "mediaTypeOverride": media_type_override.value
                    if media_type_override is not None
                    else None,
                    "includeAttachments": _params.get("attachments", False),
                    "includeEmbeddings": _params.get("embeddings", False),
                    "includeMetadata": _params.get("metadata_fields", False),
                    "includeDataRowDetails": _params.get(
                        "data_row_details", False
                    ),
                    "includeProjectDetails": _params.get(
                        "project_details", False
                    ),
                    "includePerformanceDetails": _params.get(
                        "performance_details", False
                    ),
                    "includeLabelDetails": _params.get("label_details", False),
                    "includeInterpolatedFrames": _params.get(
                        "interpolated_frames", False
                    ),
                    "includePredictions": _params.get("predictions", False),
                    "projectIds": _params.get("project_ids", None),
                    "modelRunIds": _params.get("model_run_ids", None),
                    "allProjects": _params.get("all_projects", False),
                    "allModelRuns": _params.get("all_model_runs", False),
                },
                "streamable": streamable,
            }
        }

        search_query = build_filters(self.client, _filters)
        search_query.append(
            {"ids": [self.uid], "operator": "is", "type": "dataset"}
        )

        query_params["input"]["filters"]["searchQuery"]["query"] = search_query

        res = self.client.execute(
            create_task_query_str, query_params, error_log_key="errors"
        )
        res = res[mutation_name]
        task_id = res["taskId"]
        is_streamable = res["isStreamable"]
        return Task.get_task(self.client, task_id), is_streamable

    def upsert_data_rows(
        self, items, file_upload_thread_count=FILE_UPLOAD_THREAD_COUNT
    ) -> "DataUpsertTask":
        """
        Upserts data rows in this dataset. When "key" is provided, and it references an existing data row,
        an update will be performed. When "key" is not provided a new data row will be created.

        >>>     task = dataset.upsert_data_rows([
        >>>         # create new data row
        >>>         {
        >>>             "row_data": "http://my_site.com/photos/img_01.jpg",
        >>>             "global_key": "global_key1",
        >>>             "external_id": "ex_id1",
        >>>             "attachments": [
        >>>                 {"type": AttachmentType.RAW_TEXT, "name": "att1", "value": "test1"}
        >>>             ],
        >>>             "metadata": [
        >>>                 {"name": "tag", "value": "tag value"},
        >>>             ]
        >>>         },
        >>>         # update global key of data row by existing global key
        >>>         {
        >>>             "key": GlobalKey("global_key1"),
        >>>             "global_key": "global_key1_updated"
        >>>         },
        >>>         # update data row by ID
        >>>         {
        >>>             "key": UniqueId(dr.uid),
        >>>             "external_id": "ex_id1_updated"
        >>>         },
        >>>     ])
        >>>     task.wait_till_done()
        """
        specs = DataRowUpsertItem.build(self.uid, items, (UniqueId, GlobalKey))
        return self._exec_upsert_data_rows(specs, file_upload_thread_count)

    def _exec_upsert_data_rows(
        self,
        specs: List[DataRowItemBase],
        file_upload_thread_count: int = FILE_UPLOAD_THREAD_COUNT,
    ) -> "DataUpsertTask":
        manifest = data_row_uploader.upload_in_chunks(
            client=self.client,
            specs=specs,
            file_upload_thread_count=file_upload_thread_count,
            max_chunk_size_bytes=UPSERT_CHUNK_SIZE_BYTES,
        )

        data = json.dumps(manifest.model_dump()).encode("utf-8")
        manifest_uri = self.client.upload_data(
            data, content_type="application/json", filename="manifest.json"
        )

        query_str = """
            mutation UpsertDataRowsPyApi($manifestUri: String!) {
                upsertDataRows(data: { manifestUri: $manifestUri }) { 
                    id createdAt updatedAt name status completionPercentage result errors type metadata 
                }
            }
            """

        res = self.client.execute(query_str, {"manifestUri": manifest_uri})
        res = res["upsertDataRows"]
        task = DataUpsertTask(self.client, res)
        task._user = self.client.get_user()
        return task

    def add_iam_integration(
        self, iam_integration: Union[str, IAMIntegration]
    ) -> IAMIntegration:
        """
            Sets the IAM integration for the dataset. IAM integration is used to sign URLs for data row assets.

        Args:
                iam_integration (Union[str, IAMIntegration]): IAM integration object or IAM integration id.

            Returns:
                IAMIntegration: IAM integration object.

            Raises:
                LabelboxError: If the IAM integration can't be set.

            Examples:

                >>>    # Get all IAM integrations
                >>>    iam_integrations = client.get_organization().get_iam_integrations()
                >>>
                >>>    # Get IAM integration id
                >>>    iam_integration_id = [integration.uid for integration
                >>>      in iam_integrations
                >>>      if integration.name == "My S3 integration"][0]
                >>>
                >>>   # Set IAM integration for integration id
                >>>   dataset.set_iam_integration(iam_integration_id)
                >>>
                >>>    # Get IAM integration object
                >>>    iam_integration = [integration.uid for integration
                >>>      in iam_integrations
                >>>      if integration.name == "My S3 integration"][0]
                >>>
                >>>   # Set IAM integration for IAMIntegrtion object
                >>>   dataset.set_iam_integration(iam_integration)
        """

        iam_integration_id = (
            iam_integration.uid
            if isinstance(iam_integration, IAMIntegration)
            else iam_integration
        )

        query = """
            mutation SetSignerForDatasetPyApi($signerId: ID!, $datasetId: ID!) {
                setSignerForDataset(
                    data: { signerId: $signerId }
                    where: { id: $datasetId }
                ) {
                    id
                    signer {
                        id
                    }
                }
            }
        """

        response = self.client.execute(
            query, {"signerId": iam_integration_id, "datasetId": self.uid}
        )

        if not response:
            raise ResourceNotFoundError(
                IAMIntegration,
                {"signerId": iam_integration_id, "datasetId": self.uid},
            )

        try:
            iam_integration_id = response.get("setSignerForDataset", {}).get(
                "signer", {}
            )["id"]

            return [
                integration
                for integration in self.client.get_organization().get_iam_integrations()
                if integration.uid == iam_integration_id
            ][0]
        except:
            raise LabelboxError(
                f"Can't retrieve IAM integration {iam_integration_id}"
            )

    def remove_iam_integration(self) -> None:
        """
        Unsets the IAM integration for the dataset.

        Args:
            None

        Returns:
            None

        Raises:
            LabelboxError: If the IAM integration can't be unset.

        Examples:
            >>> dataset.remove_iam_integration()
        """

        query = """
        mutation DetachSignerPyApi($id: ID!) {
            clearSignerForDataset(where: { id: $id }) {
                id
            }
        }
        """

        response = self.client.execute(query, {"id": self.uid})

        if not response:
            raise ResourceNotFoundError(Dataset, {"id": self.uid})



import json
import warnings
from dataclasses import dataclass
from io import BytesIO
from typing import List, Optional

import requests
from lbox.exceptions import (
    InternalServerError,
    LabelboxError,
    ResourceNotFoundError,
)

from labelbox import Client
from labelbox.pagination import PaginatedCollection


@dataclass
class UploadReportLine:
    """A single line in the CSV report of the upload members mutation.
    Both errors and successes are reported here.

    Example output when using dataclasses.asdict():
    >>> {
    >>>     'lines': [
    >>>         {
    >>>             'email': '...',
    >>>             'result': 'Not added',
    >>>             'error': 'User not found in the current organization'
    >>>         },
    >>>         {
    >>>             'email': '...',
    >>>             'result': 'Not added',
    >>>             'error': 'Member already exists in group'
    >>>         },
    >>>         {
    >>>             'email': '...',
    >>>             'result': 'Added',
    >>>             'error': ''
    >>>         }
    >>>     ]
    >>> }
    """

    email: str
    result: str
    error: Optional[str] = None


@dataclass
class UploadReport:
    """The report of the upload members mutation."""

    lines: List[UploadReportLine]


@dataclass
class Member:
    """A member of a user group."""

    email: str


class UserGroupV2:
    """Upload members to a user group."""

    def __init__(self, client: Client):
        self.client = client

    def upload_members(
        self, group_id: str, role: str, emails: List[str]
    ) -> Optional[UploadReport]:
        """Upload members to a user group.

        Args:
            group_id: A valid ID of the user group.
            role: The name of the role to assign to the uploaded members as it appears in the UI on the Import Members popup.
            emails: The list of emails of the members to upload.

        Returns:
            UploadReport: The report of the upload members mutation.

        Raises:
            ResourceNotFoundError: If the role is not found.
            LabelboxError: If the upload fails.

            For indicvidual email errors, the error message is available in the UploadReport.
        """
        warnings.warn(
            "The upload_members for UserGroupV2 is in beta. The method name and signature may change in the future.”",
        )

        if len(emails) == 0:
            print("No emails to upload.")
            return None

        role_id = self._get_role_id(role)
        if role_id is None:
            raise ResourceNotFoundError(
                message="Could not find a valid role with the name provided. Please make sure the role name is correct."
            )

        buffer = BytesIO()
        buffer.write(b"email\n")  # Header row
        for email in emails:
            buffer.write(f"{email}\n".encode("utf-8"))
        # Reset pointer to start of stream
        buffer.seek(0)

        multipart_file_field = "1"
        gql_file_field = "file"
        files = {
            multipart_file_field: (
                f"{multipart_file_field}.csv",
                buffer,
                "text/csv",
            )
        }
        query = """mutation ImportMembersToGroupPyPi(
                    $roleId: ID!
                    $file: Upload!
                    $where: WhereUniqueIdInput!
                    ) {
                    importUsersAsCsvToGroup(roleId: $roleId, file: $file, where: $where) {
                        csvReport
                        addedCount
                        count
                    }
                }
            """
        params = {
            "roleId": role_id,
            gql_file_field: None,
            "where": {"id": group_id},
        }

        request_data = {
            "operations": json.dumps(
                {
                    "variables": params,
                    "query": query,
                }
            ),
            "map": (
                None,
                json.dumps(
                    {multipart_file_field: [f"variables.{gql_file_field}"]}
                ),
            ),
        }

        client = self.client
        headers = dict(client.connection.headers)
        headers.pop("Content-Type", None)
        request = requests.Request(
            "POST",
            client.endpoint,
            headers=headers,
            data=request_data,
            files=files,
        )

        prepped: requests.PreparedRequest = request.prepare()

        response = client.connection.send(prepped)

        if response.status_code == 502:
            error_502 = "502 Bad Gateway"
            raise InternalServerError(error_502)
        elif response.status_code == 503:
            raise InternalServerError(response.text)
        elif response.status_code == 520:
            raise InternalServerError(response.text)

        try:
            file_data = response.json().get("data", None)
        except ValueError as e:  # response is not valid JSON
            raise LabelboxError("Failed to upload, unknown cause", e)

        if not file_data or not file_data.get("importUsersAsCsvToGroup", None):
            try:
                errors = response.json().get("errors", [])
                error_msg = "Unknown error"
                if errors:
                    error_msg = errors[0].get("message", "Unknown error")
            except Exception:
                error_msg = "Unknown error"
            raise LabelboxError("Failed to upload, message: %s" % error_msg)

        csv_report = file_data["importUsersAsCsvToGroup"]["csvReport"]
        return self._parse_csv_report(csv_report)

    def export_members(self, group_id: str) -> Optional[List[Member]]:
        warnings.warn(
            "The export_members for UserGroupV2 is in beta. The method name and signature may change in the future.",
        )

        if not group_id:
            raise ValueError("Group id is required")

        query = """query GetExportMembersAsCSVPyPi(
            $id: ID!
            ) {
            userGroupV2(where: { id: $id }) {
                id
                membersAsCSV
            }
        }
        """
        params = {
            "id": group_id,
        }

        result = self.client.execute(query, params)
        if result["userGroupV2"] is None:
            raise ResourceNotFoundError(message="The user group is not found.")
        data = result["userGroupV2"]

        return self._parse_members_csv(data["membersAsCSV"])

    def _parse_members_csv(self, csv_data: str) -> List[Member]:
        csv_lines = csv_data.strip().split("\n")
        if not csv_lines:
            return []

        members_list = []
        # Skip header row
        for email in csv_lines[1:]:
            if email.strip():  # Skip empty lines
                members_list.append(Member(email=email.strip()))

        return members_list

    def _get_role_id(self, role_name: str) -> Optional[str]:
        role_id = None
        query = """query GetAvailableUserRolesPyPi {
                    roles(skip: %d, first: %d) {
                        id
                        organizationId
                        name
                        description
                    }
                }
            """

        result = PaginatedCollection(
            client=self.client,
            query=query,
            params={},
            dereferencing=["roles"],
            obj_class=lambda _, data: data,  # type: ignore
        )
        if result is None:
            raise ResourceNotFoundError(
                message="Could not find any valid roles."
            )
        for role in result:
            if role["name"].strip() == role_name.strip():
                role_id = role["id"]
                break

        return role_id

    def _parse_csv_report(self, csv_report: str) -> UploadReport:
        lines = csv_report.strip().split("\n")
        headers = lines[0].split(",")
        report_lines = []
        for line in lines[1:]:
            values = line.split(",")
            row = dict(zip(headers, values))
            report_lines.append(
                UploadReportLine(
                    email=row["Email"],
                    result=row["Result"],
                    error=row.get(
                        "Error"
                    ),  # Using get() since error is optional
                )
            )
        return UploadReport(lines=report_lines)



import json
import random

import pytest

from labelbox.schema.data_row_payload_templates import ModelEvaluationTemplate


@pytest.fixture
def mmc_data_row(dataset):
    data = ModelEvaluationTemplate()

    content_all = data.model_dump(exclude_none=True)
    task = dataset.create_data_rows([content_all])
    task.wait_till_done()
    assert task.status == "COMPLETE"

    data_row = list(dataset.data_rows())[0]

    yield data_row

    data_row.delete()


@pytest.fixture
def mmc_data_row_all(dataset, make_metadata_fields, embedding, rand_gen):
    data = ModelEvaluationTemplate()
    data.row_data.rootMessageIds = ["root1"]
    global_key = f"global_key_{rand_gen(str)}"
    data.global_key = global_key
    vector = [random.uniform(1.0, 2.0) for _ in range(embedding.dims)]
    data.embeddings = [{"embedding_id": embedding.id, "vector": vector}]
    data.metadata_fields = make_metadata_fields
    data.attachments = [{"type": "RAW_TEXT", "value": "attachment value"}]

    content_all = data.model_dump(exclude_none=True)
    task = dataset.create_data_rows([content_all])
    task.wait_till_done()
    assert task.status == "COMPLETE"

    data_row = list(dataset.data_rows())[0]

    yield data_row, global_key

    data_row.delete()


def test_mmc(mmc_data_row):
    data_row = mmc_data_row
    assert json.loads(data_row.row_data) == {
        "type": "application/vnd.labelbox.conversational.model-chat-evaluation",
        "draft": True,
        "rootMessageIds": [],
        "actors": {},
        "messages": {},
        "version": 2,
    }


def test_mmc_all(mmc_data_row_all, embedding, constants):
    data_row, global_key = mmc_data_row_all
    assert json.loads(data_row.row_data) == {
        "type": "application/vnd.labelbox.conversational.model-chat-evaluation",
        "draft": True,
        "rootMessageIds": ["root1"],
        "actors": {},
        "messages": {},
        "version": 2,
    }
    assert data_row.global_key == global_key
    metadata_fields = data_row.metadata_fields
    metadata = data_row.metadata
    assert len(metadata_fields) == 3
    assert len(metadata) == 3
    assert [m["schemaId"] for m in metadata_fields].sort() == constants[
        "EXPECTED_METADATA_SCHEMA_IDS"
    ].sort()

    attachments = list(data_row.attachments())
    assert len(attachments) == 1

    assert embedding.get_imported_vector_count() == 1



from typing import TYPE_CHECKING, Dict, List, Optional

from lbox.exceptions import LabelboxError

from labelbox import utils
from labelbox.orm.db_object import DbObject, Entity, query
from labelbox.orm.model import Field, Relationship
from labelbox.schema.invite import InviteLimit
from labelbox.schema.resource_tag import ResourceTag

if TYPE_CHECKING:
    from labelbox import (
        IAMIntegration,
        Invite,
        InviteLimit,
        ProjectRole,
        Role,
        User,
    )


class Organization(DbObject):
    """An Organization is a group of Users.

    It is associated with data created by Users within that Organization.
    Typically all Users within an Organization have access to data created by any User in the same Organization.

    Attributes:
        updated_at (datetime)
        created_at (datetime)
        name (str)

        users (Relationship): `ToMany` relationship to User
        projects (Relationship): `ToMany` relationship to Project
        webhooks (Relationship): `ToMany` relationship to Webhook
    """

    # RelationshipManagers in Organization use the type in Query (and
    # not the source object) because the server-side does not support
    # filtering on ID in the query for getting a single organization.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for relationship in self.relationships():
            getattr(self, relationship.name).filter_on_id = False

    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    name = Field.String("name")

    # Relationships
    users = Relationship.ToMany("User", False)
    projects = Relationship.ToMany("Project", True)
    webhooks = Relationship.ToMany("Webhook", False)
    resource_tags = Relationship.ToMany("ResourceTags", False)

    def invite_user(
        self,
        email: str,
        role: "Role",
        project_roles: Optional[List["ProjectRole"]] = None,
    ) -> "Invite":
        """
        Invite a new member to the org. This will send the user an email invite

        Args:
            email (str): email address of the user to invite
            role (Role): Role to assign to the user
            project_roles (Optional[List[ProjectRoles]]): List of project roles to assign to the User (if they have a project based org role).

        Returns:
            Invite for the user

        Notes:
            1. Multiple invites can be sent for the same email. This can only be resolved in the UI for now.
                - Future releases of the SDK will support the ability to query and revoke invites to solve this problem (and/or checking on the backend)
            2. Some server side response are unclear (e.g. if the user invites themself `None` is returned which the SDK raises as a `LabelboxError` )
        """

        if project_roles and role.name != "NONE":
            raise ValueError(
                f"Project roles cannot be set for a user with organization level permissions. Found role name `{role.name}`, expected `NONE`"
            )

        data_param = "data"
        query_str = """mutation createInvitesPyApi($%s: [CreateInviteInput!]){
                    createInvites(data: $%s){  invite { id createdAt organizationRoleName inviteeEmail inviter { %s } }}}""" % (
            data_param,
            data_param,
            query.results_query_part(Entity.User),
        )

        projects = [
            {
                "projectId": project_role.project.uid,
                "projectRoleId": project_role.role.uid,
            }
            for project_role in project_roles or []
        ]

        res = self.client.execute(
            query_str,
            {
                data_param: [
                    {
                        "inviterId": self.client.get_user().uid,
                        "inviteeEmail": email,
                        "organizationId": self.uid,
                        "organizationRoleId": role.uid,
                        "projects": projects,
                    }
                ]
            },
        )
        invite_response = res["createInvites"][0]["invite"]
        if not invite_response:
            raise LabelboxError(f"Unable to send invite for email {email}")
        return Entity.Invite(self.client, invite_response)

    def invite_limit(self) -> InviteLimit:
        """Retrieve invite limits for the org
        This already accounts for users currently in the org
        Meaining that  `used = users + invites, remaining = limit - (users + invites)`

        Returns:
            InviteLimit

        """
        org_id_param = "organizationId"
        res = self.client.execute(
            """query InvitesLimitPyApi($%s: ID!) {
            invitesLimit(where: {id: $%s}) { used limit remaining }
        }"""
            % (org_id_param, org_id_param),
            {org_id_param: self.uid},
        )
        return InviteLimit(
            **{utils.snake_case(k): v for k, v in res["invitesLimit"].items()}
        )

    def remove_user(self, user: "User") -> None:
        """
        Deletes a user from the organization. This cannot be undone without sending another invite.

        Args:
            user (User): The user to delete from the org
        """

        user_id_param = "userId"
        self.client.execute(
            """mutation DeleteMemberPyApi($%s: ID!) {
            updateUser(where: {id: $%s}, data: {deleted: true}) { id deleted }
        }"""
            % (user_id_param, user_id_param),
            {user_id_param: user.uid},
        )

    def create_resource_tag(self, tag: Dict[str, str]) -> ResourceTag:
        """
        Creates a resource tag.
            >>> tag = {'text': 'tag-1',  'color': 'ffffff'}

        Args:
            tag (dict): A resource tag {'text': 'tag-1', 'color': 'fffff'}
        Returns:
            The created resource tag.
        """
        tag_text_param = "text"
        tag_color_param = "color"

        query_str = """mutation CreateResourceTagPyApi($text:String!,$color:String!) {
                createResourceTag(input:{text:$%s,color:$%s}) {%s}}
        """ % (
            tag_text_param,
            tag_color_param,
            query.results_query_part(ResourceTag),
        )

        params = {
            tag_text_param: tag.get("text", None),
            tag_color_param: tag.get("color", None),
        }
        if not all(params.values()):
            raise ValueError(
                f"tag must contain 'text' and 'color' keys. received: {tag}"
            )

        res = self.client.execute(query_str, params)
        return ResourceTag(self.client, res["createResourceTag"])

    def get_resource_tags(self) -> List[ResourceTag]:
        """
        Returns all resource tags for an organization
        """
        query_str = (
            """query GetOrganizationResourceTagsPyApi{organization{resourceTag{%s}}}"""
            % (query.results_query_part(ResourceTag))
        )

        return [
            ResourceTag(self.client, tag)
            for tag in self.client.execute(query_str)["organization"][
                "resourceTag"
            ]
        ]

    def get_iam_integrations(self) -> List["IAMIntegration"]:
        """
        Returns all IAM Integrations for an organization
        """
        res = self.client.execute(
            """query getAllIntegrationsPyApi { iamIntegrations {
                %s
                settings {
                __typename
                ... on AwsIamIntegrationSettings {roleArn}
                ... on GcpIamIntegrationSettings {serviceAccountEmailId readBucket}
                }

            } } """
            % query.results_query_part(Entity.IAMIntegration)
        )
        return [
            Entity.IAMIntegration(self.client, integration_data)
            for integration_data in res["iamIntegrations"]
        ]

    def get_default_iam_integration(self) -> Optional["IAMIntegration"]:
        """
        Returns the default IAM integration for the organization.
        Will return None if there are no default integrations for the org.
        """
        integrations = self.get_iam_integrations()
        default_integration = [
            integration
            for integration in integrations
            if integration.is_org_default
        ]
        if len(default_integration) > 1:
            raise ValueError(
                "Found more than one default signer. Please contact Labelbox to resolve"
            )
        return (
            None if not len(default_integration) else default_integration.pop()
        )



from unittest.mock import MagicMock

import pytest
import requests
from lbox.exceptions import ResourceCreationError, ResourceNotFoundError

from labelbox import Dataset
from labelbox.schema.internal.descriptor_file_creator import (
    DescriptorFileCreator,
)


def test_dataset(client, rand_gen):
    # confirm dataset can be created
    name = rand_gen(str)
    dataset = client.create_dataset(name=name)
    assert dataset.name == name
    assert dataset.created_by() == client.get_user()
    assert dataset.organization() == client.get_organization()

    retrieved_dataset = client.get_dataset(dataset.uid)
    assert retrieved_dataset.name == dataset.name
    assert retrieved_dataset.uid == dataset.uid
    assert retrieved_dataset.created_by() == dataset.created_by()
    assert retrieved_dataset.organization() == dataset.organization()

    dataset = client.get_dataset(dataset.uid)
    assert dataset.name == name

    new_name = rand_gen(str)
    dataset.update(name=new_name)
    # Test local object updates.
    assert dataset.name == new_name

    # Test remote updates.
    dataset = client.get_dataset(dataset.uid)
    assert dataset.name == new_name

    # Test description
    description = rand_gen(str)
    assert dataset.description == ""
    dataset.update(description=description)
    assert dataset.description == description

    dataset.delete()

    with pytest.raises(ResourceNotFoundError):
        dataset = client.get_dataset(dataset.uid)


@pytest.fixture
def dataset_for_filtering(client, rand_gen):
    name_1 = rand_gen(str)
    name_2 = rand_gen(str)
    d1 = client.create_dataset(name=name_1)
    d2 = client.create_dataset(name=name_2)

    yield name_1, d1, name_2, d2


def test_dataset_filtering(client, dataset_for_filtering):
    name_1, d1, name_2, d2 = dataset_for_filtering

    assert list(client.get_datasets(where=Dataset.name == name_1)) == [d1]
    assert list(client.get_datasets(where=Dataset.name == name_2)) == [d2]


def test_dataset_filtering(client, dataset_for_filtering):
    name_1, d1, name_2, d2 = dataset_for_filtering

    assert list(client.get_datasets(where=Dataset.name == name_1)) == [d1]
    assert list(client.get_datasets(where=Dataset.name == name_2)) == [d2]


def test_get_data_row_for_external_id(dataset, rand_gen, image_url):
    external_id = rand_gen(str)

    with pytest.raises(ResourceNotFoundError):
        data_row = dataset.data_row_for_external_id(external_id)

    data_row = dataset.create_data_row(
        row_data=image_url, external_id=external_id
    )

    found = dataset.data_row_for_external_id(external_id)
    assert found.uid == data_row.uid
    assert found.external_id == external_id

    dataset.create_data_row(row_data=image_url, external_id=external_id)
    assert len(dataset.data_rows_for_external_id(external_id)) == 2

    task = dataset.create_data_rows(
        [dict(row_data=image_url, external_id=external_id)]
    )
    task.wait_till_done()
    assert len(dataset.data_rows_for_external_id(external_id)) == 3


def test_upload_video_file(dataset, sample_video: str) -> None:
    """
    Tests that a mp4 video can be uploaded and preserve content length
    and content type.

    """
    dataset.create_data_row(row_data=sample_video)
    task = dataset.create_data_rows([sample_video, sample_video])
    task.wait_till_done()

    with open(sample_video, "rb") as video_f:
        content_length = len(video_f.read())

    for data_row in dataset.data_rows():
        url = data_row.row_data
        response = requests.head(url, allow_redirects=True)
        assert int(response.headers["Content-Length"]) == content_length
        assert response.headers["Content-Type"] == "video/mp4"


def test_create_pdf(dataset):
    dataset.create_data_row(
        row_data={
            "pdfUrl": "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/sample-document-1.pdf",
            "textLayerUrl": "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/sample-document-custom-text-layer.json",
        }
    )
    dataset.create_data_row(
        row_data={
            "pdfUrl": "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/sample-document-1.pdf",
            "textLayerUrl": "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/sample-document-custom-text-layer.json",
        },
        media_type="PDF",
    )

    with pytest.raises(ResourceCreationError):
        # Wrong media type
        dataset.create_data_row(
            row_data={
                "pdfUrl": "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/sample-document-1.pdf",
                "textLayerUrl": "https://lb-test-data.s3.us-west-1.amazonaws.com/document-samples/sample-document-custom-text-layer.json",
            },
            media_type="TEXT",
        )


def test_bulk_conversation(dataset, sample_bulk_conversation: list) -> None:
    """
    Tests that bulk conversations can be uploaded.

    """
    task = dataset.create_data_rows(sample_bulk_conversation)
    task.wait_till_done()

    assert len(list(dataset.data_rows())) == len(sample_bulk_conversation)


def test_create_descriptor_file(dataset):
    import unittest.mock as mock

    client = MagicMock()
    with mock.patch.object(
        client, "upload_data", wraps=client.upload_data
    ) as upload_data_spy:
        DescriptorFileCreator(client).create_one(
            items=[{"row_data": "some text..."}]
        )
        upload_data_spy.assert_called()
        call_args, call_kwargs = (
            upload_data_spy.call_args_list[0][0],
            upload_data_spy.call_args_list[0][1],
        )
        assert call_args == ('[{"row_data": "some text..."}]',)
        assert call_kwargs == {
            "content_type": "application/json",
            "filename": "json_import.json",
        }



# type: ignore

import colorsys
import json
from dataclasses import dataclass, field
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from lbox.exceptions import InconsistentOntologyException

from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field, Relationship
from labelbox.schema.tool_building.classification import (
    Classification,
    PromptResponseClassification,
)
from labelbox.schema.tool_building.fact_checking_tool import (
    FactCheckingTool,
)
from labelbox.schema.tool_building.prompt_issue_tool import PromptIssueTool
from labelbox.schema.tool_building.step_reasoning_tool import (
    StepReasoningTool,
)
from labelbox.schema.tool_building.tool_type import ToolType
from labelbox.schema.tool_building.tool_type_mapping import (
    map_tool_type_to_tool_cls,
)
from labelbox.schema.tool_building.types import (
    FeatureSchemaAttribute,
    FeatureSchemaAttributes,
)
import warnings


class DeleteFeatureFromOntologyResult:
    archived: bool
    deleted: bool

    def __str__(self):
        return "<%s %s>" % (
            self.__class__.__name__.split(".")[-1],
            json.dumps(self.__dict__),
        )


class FeatureSchema(DbObject):
    name = Field.String("name")
    color = Field.String("name")
    normalized = Field.Json("normalized")


@dataclass
class Tool:
    """
    A tool to be added to a Project's ontology. The tool is
    dependent on the Tool Type.

    To instantiate, the "tool" and "name" parameters must
    be passed in.

    The "classifications" parameter holds a list of Classification objects.
    This can be used to add nested classifications to a tool.

    Example(s):
        tool = Tool(
            tool = Tool.Type.LINE,
            name = "Tool example")
        classification = Classification(
            class_type = Classification.Type.TEXT,
            instructions = "Classification Example")
        tool.add_classification(classification)

    Attributes:
        tool: (Tool.Type)
        name: (str)
        required: (bool)
        color: (str)
        classifications: (list)
        schema_id: (str)
        feature_schema_id: (str)
        attributes: (list)
    """

    class Type(Enum):
        POLYGON = "polygon"
        SEGMENTATION = "superpixel"
        RASTER_SEGMENTATION = "raster-segmentation"
        POINT = "point"
        BBOX = "rectangle"
        LINE = "line"
        NER = "named-entity"
        RELATIONSHIP = "edge"
        MESSAGE_SINGLE_SELECTION = "message-single-selection"
        MESSAGE_MULTI_SELECTION = "message-multi-selection"
        MESSAGE_RANKING = "message-ranking"

    tool: Type
    name: str
    required: bool = False
    color: Optional[str] = None
    classifications: List[Classification] = field(default_factory=list)
    schema_id: Optional[str] = None
    feature_schema_id: Optional[str] = None
    attributes: Optional[FeatureSchemaAttributes] = None

    def __post_init__(self):
        if self.attributes is not None:
            warnings.warn(
                "The attributes for Tools are in beta. The attribute name and signature may change in the future."
            )

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> Dict[str, Any]:
        return cls(
            name=dictionary["name"],
            schema_id=dictionary.get("schemaNodeId", None),
            feature_schema_id=dictionary.get("featureSchemaId", None),
            required=dictionary.get("required", False),
            tool=Tool.Type(dictionary["tool"]),
            classifications=[
                Classification.from_dict(c)
                for c in dictionary["classifications"]
            ],
            color=dictionary["color"],
            attributes=[
                FeatureSchemaAttribute.from_dict(attr)
                for attr in dictionary.get("attributes", []) or []
            ]
            if dictionary.get("attributes")
            else None,
        )

    def asdict(self) -> Dict[str, Any]:
        return {
            "tool": self.tool.value,
            "name": self.name,
            "required": self.required,
            "color": self.color,
            "classifications": [
                c.asdict(is_subclass=True) for c in self.classifications
            ],
            "schemaNodeId": self.schema_id,
            "featureSchemaId": self.feature_schema_id,
            "attributes": [a.asdict() for a in self.attributes]
            if self.attributes is not None
            else None,
        }

    def add_classification(self, classification: Classification) -> None:
        if classification.name in (c.name for c in self.classifications):
            raise InconsistentOntologyException(
                f"Duplicate nested classification '{classification.name}' "
                f"for tool '{self.name}'"
            )
        self.classifications.append(classification)


"""
The following 2 functions help to bridge the gap between the step reasoning all other tool ontologies.
"""


def tool_cls_from_type(tool_type: str):
    tool_cls = map_tool_type_to_tool_cls(tool_type)
    if tool_cls is not None:
        return tool_cls
    return Tool


def tool_type_cls_from_type(tool_type: str):
    if ToolType.valid(tool_type):
        return ToolType
    return Tool.Type


class Ontology(DbObject):
    """An ontology specifies which tools and classifications are available
    to a project. This is read only for now.
    Attributes:
        name (str)
        description (str)
        updated_at (datetime)
        created_at (datetime)
        normalized (json)
        object_schema_count (int)
        classification_schema_count (int)
        projects (Relationship): `ToMany` relationship to Project
        created_by (Relationship): `ToOne` relationship to User
    """

    name = Field.String("name")
    description = Field.String("description")
    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    normalized = Field.Json("normalized")
    object_schema_count = Field.Int("object_schema_count")
    classification_schema_count = Field.Int("classification_schema_count")

    projects = Relationship.ToMany("Project", True)
    created_by = Relationship.ToOne("User", False, "created_by")

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._tools: Optional[List[Tool]] = None
        self._classifications: Optional[
            Union[List[Classification], List[PromptResponseClassification]]
        ] = None

    def tools(self) -> List[Tool]:
        """Get list of tools (AKA objects) in an Ontology."""
        if self._tools is None:
            self._tools = [
                tool_cls_from_type(tool["tool"]).from_dict(tool)
                for tool in self.normalized["tools"]
            ]
        return self._tools

    def classifications(
        self,
    ) -> List[Union[Classification, PromptResponseClassification]]:
        """Get list of classifications in an Ontology."""
        if self._classifications is None:
            self._classifications = []
            for classification in self.normalized["classifications"]:
                if (
                    "type" in classification
                    and classification["type"]
                    in PromptResponseClassification.Type._value2member_map_.keys()
                ):
                    self._classifications.append(
                        PromptResponseClassification.from_dict(classification)
                    )
                else:
                    self._classifications.append(
                        Classification.from_dict(classification)
                    )
        return self._classifications


@dataclass
class OntologyBuilder:
    """
    A class to help create an ontology for a Project. This should be used
    for making Project ontologies from scratch. OntologyBuilder can also
    pull from an already existing Project's ontology.

    There are no required instantiation arguments.

    To create an ontology, use the asdict() method after fully building your
    ontology within this class, and inserting it into client.create_ontology() as the
    "normalized" parameter.

    Example:
        >>> builder = OntologyBuilder()
        >>> ...
        >>> ontology = client.create_ontology(
        >>>    "Ontology from new features",
        >>>    ontology_builder.asdict(),
        >>>    media_type=lb.MediaType.Image,
        >>> )
        >>> project.connect_ontology(ontology)

    attributes:
        tools: (list)
        classifications: (list)


    """

    tools: List[
        Union[Tool, StepReasoningTool, FactCheckingTool, PromptIssueTool]
    ] = field(default_factory=list)
    classifications: List[
        Union[Classification, PromptResponseClassification]
    ] = field(default_factory=list)

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> Dict[str, Any]:
        classifications = []
        for c in dictionary["classifications"]:
            if (
                "type" in c
                and c["type"]
                in PromptResponseClassification.Type._value2member_map_.keys()
            ):
                classifications.append(
                    PromptResponseClassification.from_dict(c)
                )
            else:
                classifications.append(Classification.from_dict(c))
        return cls(
            tools=[Tool.from_dict(t) for t in dictionary["tools"]],
            classifications=classifications,
        )

    def asdict(self) -> Dict[str, Any]:
        self._update_colors()
        classifications = []
        prompts = 0
        for c in self.classifications:
            if (
                hasattr(c, "class_type")
                and c.class_type in PromptResponseClassification.Type
            ):
                if c.class_type == PromptResponseClassification.Type.PROMPT:
                    prompts += 1
                    if prompts > 1:
                        raise ValueError(
                            "Only one prompt is allowed per ontology"
                        )
                classifications.append(PromptResponseClassification.asdict(c))
            else:
                classifications.append(Classification.asdict(c))
        return {
            "tools": [t.asdict() for t in self.tools],
            "classifications": classifications,
        }

    def _update_colors(self):
        num_tools = len(self.tools)

        for index in range(num_tools):
            hsv_color = (index * 1 / num_tools, 1, 1)
            rgb_color = tuple(
                int(255 * x) for x in colorsys.hsv_to_rgb(*hsv_color)
            )
            if self.tools[index].color is None:
                self.tools[index].color = "#%02x%02x%02x" % rgb_color

    @classmethod
    def from_project(cls, project: "project.Project") -> "OntologyBuilder":
        ontology = project.ontology().normalized
        return cls.from_dict(ontology)

    @classmethod
    def from_ontology(cls, ontology: Ontology) -> "OntologyBuilder":
        return cls.from_dict(ontology.normalized)

    def add_tool(self, tool: Tool) -> None:
        if tool.name in (t.name for t in self.tools):
            raise InconsistentOntologyException(
                f"Duplicate tool name '{tool.name}'. "
            )
        self.tools.append(tool)

    def add_classification(
        self,
        classification: Union[Classification, PromptResponseClassification],
    ) -> None:
        if classification.name in (c.name for c in self.classifications):
            raise InconsistentOntologyException(
                f"Duplicate classification name '{classification.name}'. "
            )
        self.classifications.append(classification)



import pytest
from unittest.mock import MagicMock

from labelbox.schema.data_row import DataRow
from labelbox.schema.identifiable import GlobalKey, UniqueId
from labelbox.schema.project import validate_labeling_parameter_overrides


def test_validate_labeling_parameter_overrides_invalid_priority():
    mock_data_row = MagicMock(spec=DataRow)
    mock_data_row.uid = "abc"
    data = [
        (mock_data_row, "invalid"),
        (UniqueId("efg"), 2),
        (GlobalKey("hij"), 3),
    ]
    with pytest.raises(TypeError):
        validate_labeling_parameter_overrides(data)


def test_validate_labeling_parameter_overrides_invalid_tuple_length():
    mock_data_row = MagicMock(spec=DataRow)
    mock_data_row.uid = "abc"
    data = [
        (mock_data_row, "invalid"),
        (UniqueId("efg"), 2),
        (GlobalKey("hij")),
    ]
    with pytest.raises(TypeError):
        validate_labeling_parameter_overrides(data)



from typing import TYPE_CHECKING
from labelbox.orm import query
from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Entity, Field, Relationship

if TYPE_CHECKING:
    from labelbox import ModelRun


class Model(DbObject):
    """A model represents a program that has been trained and
    can make predictions on new data.
    Attributes:
        name (str)
        model_runs (Relationship): `ToMany` relationship to ModelRun
    """

    name = Field.String("name")
    ontology_id = Field.String("ontology_id")
    model_runs = Relationship.ToMany("ModelRun", False)

    def create_model_run(self, name, config=None) -> "ModelRun":
        """Creates a model run belonging to this model.

        Args:
            name (string): The name for the model run.
            config (json): Model run's training metadata config
        Returns:
            ModelRun, the created model run.
        """
        name_param = "name"
        config_param = "config"
        model_id_param = "modelId"
        ModelRun = Entity.ModelRun
        query_str = """mutation CreateModelRunPyApi($%s: String!, $%s: Json, $%s: ID!) {
            createModelRun(data: {name: $%s, trainingMetadata: $%s, modelId: $%s}) {%s}}""" % (
            name_param,
            config_param,
            model_id_param,
            name_param,
            config_param,
            model_id_param,
            query.results_query_part(ModelRun),
        )
        res = self.client.execute(
            query_str,
            {name_param: name, config_param: config, model_id_param: self.uid},
        )
        return ModelRun(self.client, res["createModelRun"])

    def delete(self) -> None:
        """Deletes specified model.

        Returns:
            Query execution success.
        """
        ids_param = "ids"
        query_str = """mutation DeleteModelPyApi($%s: ID!) {
            deleteModels(where: {ids: [$%s]})}""" % (ids_param, ids_param)
        self.client.execute(query_str, {ids_param: str(self.uid)})



import pytest
from lbox.exceptions import ResourceNotFoundError

from labelbox import Model


def test_model(client, configured_project, rand_gen):
    # Get all
    models = list(client.get_models())
    for m in models:
        assert isinstance(m, Model)

    # Create
    ontology = configured_project.ontology()
    data = {"name": rand_gen(str), "ontology_id": ontology.uid}
    model = client.create_model(data["name"], data["ontology_id"])
    assert model.name == data["name"]

    # Get one
    model = client.get_model(model.uid)
    assert model.name == data["name"]

    # Delete
    model.delete()
    with pytest.raises(ResourceNotFoundError):
        client.get_model(model.uid)



# Size of a single page in a paginated query.
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from lbox.call_info import call_info_as_str

if TYPE_CHECKING:
    from labelbox import Client
    from labelbox.orm.db_object import DbObject

_PAGE_SIZE = 100


class PaginatedCollection:
    """An iterable collection of database objects (Projects, Labels, etc...).

    Implements automatic (transparent to the user) paginated fetching during
    iteration. Intended for use by library internals and not by the end user.
    For a list of attributes see __init__(...) documentation. The params of
    __init__ map exactly to object attributes.
    """

    def __init__(
        self,
        client: "Client",
        query: str,
        params: Dict[str, Union[str, int]],
        dereferencing: Union[List[str], Dict[str, Any]],
        obj_class: Union[Type["DbObject"], Callable[[Any, Any], Any]],
        cursor_path: Optional[List[str]] = None,
        experimental: bool = False,
    ):
        """Creates a PaginatedCollection.

        Args:
            client (labelbox.Client): the client used for fetching data from DB.
            query (str): Base query used for pagination. It must contain two
                '%d' placeholders, the first for pagination 'skip' clause and
                the second for the 'first' clause.
            params (dict): Query parameters.
            dereferencing (iterable): An iterable of str defining the keypath
                that needs to be dereferenced in the query result in order to
                reach the paginated objects of interest.
            obj_class (type): The class of object to be instantiated with each
                dict containing db values.
            cursor_path: If not None, this is used to find the cursor
            experimental: Used to call experimental endpoints
        """
        self._fetched_all = False
        self._data: List[Dict[str, Any]] = []
        self._data_ind = 0
        self._client = client
        self._client.set_sdk_method(call_info_as_str())

        pagination_kwargs = {
            "client": self._client,
            "obj_class": obj_class,
            "dereferencing": dereferencing,
            "experimental": experimental,
            "query": query,
            "params": params,
        }

        self.paginator = (
            _CursorPagination(cursor_path, **pagination_kwargs)
            if cursor_path
            else _OffsetPagination(**pagination_kwargs)
        )

    def __iter__(self):
        self._data_ind = 0
        return self

    def __next__(self):
        if len(self._data) <= self._data_ind:
            if self._fetched_all:
                self._client.unset_sdk_method()
                raise StopIteration()

            page_data, self._fetched_all = self.paginator.get_next_page()
            self._data.extend(page_data)
            if len(page_data) == 0:
                self._client.unset_sdk_method()
                raise StopIteration()

        rval = self._data[self._data_ind]
        self._data_ind += 1
        return rval

    def get_one(self):
        """Iterates over self and returns first value
        This method is idempotent
        """
        for value in self:
            return value

    def get_many(self, n: int):
        """Iterates over self and returns first n results
        This method is idempotent

        Args:
            n (int): Number of elements to retrieve
        """
        results = []
        i = 0

        for value in self:
            if i >= n:
                break

            results.append(value)
            i += 1

        return results


class _Pagination(ABC):
    def __init__(
        self,
        client: "Client",
        obj_class: Type["DbObject"],
        dereferencing: Dict[str, Any],
        query: str,
        params: Dict[str, Any],
        experimental: bool,
    ):
        self.client = client
        self.obj_class = obj_class
        self.dereferencing = dereferencing
        self.experimental = experimental
        self.query = query
        self.params = params

    def get_page_data(self, results: Dict[str, Any]) -> List["DbObject"]:
        for deref in self.dereferencing:
            results = results[deref]

        return [self.obj_class(self.client, result) for result in results]

    @abstractmethod
    def get_next_page(self) -> Tuple[Dict[str, Any], bool]: ...


class _CursorPagination(_Pagination):
    def __init__(self, cursor_path: List[str], *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cursor_path = cursor_path
        self.next_cursor: Optional[Any] = kwargs.get("params", {}).get("from")

    def increment_page(self, results: Dict[str, Any]):
        for path in self.cursor_path:
            results = results[path]
        self.next_cursor = results

    def fetched_all(self) -> bool:
        return not self.next_cursor

    def fetch_results(self) -> Dict[str, Any]:
        page_size = self.params.get("first", _PAGE_SIZE)
        self.params.update({"from": self.next_cursor, "first": page_size})
        return self.client.execute(
            self.query, self.params, experimental=self.experimental
        )

    def get_next_page(self):
        results = self.fetch_results()
        page_data = self.get_page_data(results)
        self.increment_page(results)
        done = self.fetched_all()
        return page_data, done


class _OffsetPagination(_Pagination):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._fetched_pages = 0

    def increment_page(self):
        self._fetched_pages += 1

    def fetched_all(self, n_items: int) -> bool:
        return n_items < _PAGE_SIZE

    def fetch_results(self) -> Dict[str, Any]:
        query = self.query % (self._fetched_pages * _PAGE_SIZE, _PAGE_SIZE)
        return self.client.execute(
            query, self.params, experimental=self.experimental
        )

    def get_next_page(self):
        results = self.fetch_results()
        page_data = self.get_page_data(results)
        self.increment_page()
        done = self.fetched_all(len(page_data))
        return page_data, done



import pytest
from labelbox.data.annotation_types.geometry.polygon import Polygon
from labelbox.data.annotation_types.geometry.point import Point
from labelbox.data.annotation_types.geometry.line import Line
from labelbox.data.annotation_types.geometry.rectangle import Rectangle
from labelbox.data.annotation_types.data.tiled_image import (
    EPSG,
    TiledBounds,
    TileLayer,
    TiledImageData,
    EPSGTransformer,
)
from pydantic import ValidationError


@pytest.mark.parametrize("epsg", list(EPSG))
def test_epsg(epsg):
    assert isinstance(epsg, EPSG)


@pytest.mark.parametrize("epsg", list(EPSG))
def test_tiled_bounds(epsg):
    top_left = Point(x=0, y=0)
    bottom_right = Point(x=50, y=50)

    tiled_bounds = TiledBounds(epsg=epsg, bounds=[top_left, bottom_right])
    assert isinstance(tiled_bounds, TiledBounds)
    assert isinstance(tiled_bounds.epsg, EPSG)


@pytest.mark.parametrize("epsg", list(EPSG))
def test_tiled_bounds_same(epsg):
    single_bound = Point(x=0, y=0)
    with pytest.raises(ValidationError):
        tiled_bounds = TiledBounds(
            epsg=epsg, bounds=[single_bound, single_bound]
        )


def test_create_tiled_image_data():
    bounds_points = [Point(x=0, y=0), Point(x=5, y=5)]
    url = (
        "https://labelbox.s3-us-west-2.amazonaws.com/pathology/{z}/{x}/{y}.png"
    )
    zoom_levels = (1, 10)

    tile_layer = TileLayer(url=url, name="slippy map tile")
    tile_bounds = TiledBounds(epsg=EPSG.EPSG4326, bounds=bounds_points)
    tiled_image_data = TiledImageData(
        tile_layer=tile_layer,
        tile_bounds=tile_bounds,
        zoom_levels=zoom_levels,
        version=2,
    )
    assert isinstance(tiled_image_data, TiledImageData)
    assert tiled_image_data.tile_bounds.bounds == bounds_points
    assert tiled_image_data.tile_layer.url == url
    assert tiled_image_data.zoom_levels == zoom_levels


def test_epsg_point_projections():
    zoom = 4

    bounds_simple = TiledBounds(
        epsg=EPSG.SIMPLEPIXEL, bounds=[Point(x=0, y=0), Point(x=256, y=256)]
    )

    bounds_3857 = TiledBounds(
        epsg=EPSG.EPSG3857,
        bounds=[
            Point(x=-104.150390625, y=30.789036751261136),
            Point(x=-81.8701171875, y=45.920587344733654),
        ],
    )
    bounds_4326 = TiledBounds(
        epsg=EPSG.EPSG4326,
        bounds=[
            Point(x=-104.150390625, y=30.789036751261136),
            Point(x=-81.8701171875, y=45.920587344733654),
        ],
    )

    point = Point(x=-11016716.012685884, y=5312679.21393289)
    point_two = Point(x=-12016716.012685884, y=5212679.21393289)
    point_three = Point(x=-13016716.012685884, y=5412679.21393289)

    line = Line(points=[point, point_two, point_three])
    polygon = Polygon(points=[point, point_two, point_three])
    rectangle = Rectangle(start=point, end=point_three)

    shapes_to_test = [point, line, polygon, rectangle]

    transformer_3857_simple = EPSGTransformer.create_geo_to_pixel_transformer(
        src_epsg=EPSG.EPSG3857,
        pixel_bounds=bounds_simple,
        geo_bounds=bounds_3857,
        zoom=zoom,
    )
    transformer_3857_4326 = EPSGTransformer.create_geo_to_geo_transformer(
        src_epsg=EPSG.EPSG3857,
        tgt_epsg=EPSG.EPSG4326,
    )
    transformer_4326_simple = EPSGTransformer.create_geo_to_pixel_transformer(
        src_epsg=EPSG.EPSG4326,
        pixel_bounds=bounds_simple,
        geo_bounds=bounds_4326,
        zoom=zoom,
    )

    for shape in shapes_to_test:
        shape_simple = transformer_3857_simple(shape=shape)

        shape_4326 = transformer_3857_4326(shape=shape)

        other_simple_shape = transformer_4326_simple(shape=shape_4326)

        assert shape_simple == other_simple_shape



import pytest


def test_create_offline_chat_evaluation_project(
    rand_gen,
    offline_chat_evaluation_project,
    chat_evaluation_ontology,
    offline_conversational_data_row,
    model_config,
):
    project = offline_chat_evaluation_project
    assert project

    ontology = chat_evaluation_ontology
    project.connect_ontology(ontology)

    assert project.labeling_frontend().name == "Editor"
    assert project.ontology().name == ontology.name

    batch = project.create_batch(
        rand_gen(str),
        [offline_conversational_data_row.uid],  # sample of data row objects
    )
    assert batch

    # Can not add a model config to an offline chat evaluation project, since we do not use live models
    with pytest.raises(Exception):
        project.add_model_config(model_config.uid)



from types import SimpleNamespace
import pytest
import numpy as np
import base64


class NameSpace(SimpleNamespace):
    def __init__(
        self,
        predictions,
        labels,
        expected,
        expected_without_subclasses=None,
        data_row_expected=None,
        media_attributes=None,
        metadata=None,
        classifications=None,
    ):
        super(NameSpace, self).__init__(
            predictions=predictions,
            labels={
                "DataRow ID": "ckppihxc10005aeyjen11h7jh",
                "Labeled Data": "https://.jpg",
                "Media Attributes": media_attributes or {},
                "DataRow Metadata": metadata or [],
                "Label": {
                    "objects": labels,
                    "classifications": classifications or [],
                },
            },
            expected=expected,
            expected_without_subclasses=expected_without_subclasses or expected,
            data_row_expected=data_row_expected,
        )


@pytest.fixture
def polygon_pair():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "polygon": [
                    {"x": 0, "y": 0},
                    {"x": 1, "y": 0},
                    {"x": 1, "y": 1},
                    {"x": 0, "y": 1},
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "polygon": [
                    {"x": 0, "y": 0},
                    {"x": 1, "y": 0},
                    {"x": 1, "y": 0.5},
                    {"x": 0, "y": 0.5},
                ],
            }
        ],
        expected=0.5,
    )


@pytest.fixture
def box_pair():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "bbox": {
                    "top": 1099,
                    "left": 2010,
                    "height": 690,
                    "width": 591,
                },
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "bbox": {
                    "top": 1099,
                    "left": 2010,
                    "height": 690,
                    "width": 591,
                },
            }
        ],
        expected=1.0,
    )


@pytest.fixture
def unmatched_prediction():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "polygon": [
                    {"x": 0, "y": 0},
                    {"x": 1, "y": 0},
                    {"x": 1, "y": 1},
                    {"x": 0, "y": 1},
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "polygon": [
                    {"x": 0, "y": 0},
                    {"x": 1, "y": 0},
                    {"x": 1, "y": 0.5},
                    {"x": 0, "y": 0.5},
                ],
            },
            {
                "uuid": "d0ba2520-02e9-47d4-8736-088bbdbabbc3",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "polygon": [
                    {"x": 10, "y": 10},
                    {"x": 11, "y": 10},
                    {"x": 11, "y": 1.5},
                    {"x": 10, "y": 1.5},
                ],
            },
        ],
        expected=0.25,
    )


@pytest.fixture
def unmatched_label():
    return NameSpace(
        labels=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "polygon": [
                    {"x": 0, "y": 0},
                    {"x": 1, "y": 0},
                    {"x": 1, "y": 1},
                    {"x": 0, "y": 1},
                ],
            },
            {
                "featureId": "ckppiw3bs0007aeyjs3pvrqzi",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "polygon": [
                    {"x": 10, "y": 10},
                    {"x": 11, "y": 10},
                    {"x": 11, "y": 11},
                    {"x": 10, "y": 11},
                ],
            },
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "polygon": [
                    {"x": 0, "y": 0},
                    {"x": 1, "y": 0},
                    {"x": 1, "y": 0.5},
                    {"x": 0, "y": 0.5},
                ],
            }
        ],
        expected=0.25,
    )


def create_mask_url(indices, h, w, value):
    mask = np.zeros((h, w, 3), dtype=np.uint8)
    for idx in indices:
        mask[idx] = value
    return base64.b64encode(mask.tobytes()).decode("utf-8")


@pytest.fixture
def mask_pair():
    return NameSpace(
        labels=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "instanceURI": create_mask_url(
                    [(0, 0), (0, 1)], 32, 32, (255, 255, 255)
                ),
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "mask": {
                    "instanceURI": create_mask_url([(0, 0)], 32, 32, (1, 1, 1)),
                    "colorRGB": (1, 1, 1),
                },
            }
        ],
        expected=0.5,
    )


@pytest.fixture
def matching_radio():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckrm02no8000008l3arwp6h4f",
                "answer": {"schemaId": "ckppid25v0000aeyjmxfwlc7t"},
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckrm02no8000008l3arwp6h4f",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answer": {"schemaId": "ckppid25v0000aeyjmxfwlc7t"},
            }
        ],
        expected=1.0,
    )


@pytest.fixture
def empty_radio_label():
    return NameSpace(
        labels=[],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answer": {"schemaId": "ckppid25v0000aeyjmxfwlc7t"},
            }
        ],
        expected=0,
    )


@pytest.fixture
def empty_radio_prediction():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "answer": {"schemaId": "ckppid25v0000aeyjmxfwlc7t"},
            }
        ],
        predictions=[],
        expected=0,
    )


@pytest.fixture
def matching_checklist():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                    {"schemaId": "ckppidq4u0002aeyjmcc4toxw"},
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                    {"schemaId": "ckppidq4u0002aeyjmcc4toxw"},
                ],
            }
        ],
        data_row_expected=1.0,
        expected={1.0: 3},
    )


@pytest.fixture
def partially_matching_checklist_1():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                    {"schemaId": "ckppidq4u0002aeyjmcc4toxw"},
                    {"schemaId": "ckppie29m0003aeyjk1ixzcom"},
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                    {"schemaId": "ckppidq4u0002aeyjmcc4toxw"},
                    {"schemaId": "ckppiebx80004aeyjuwvos69e"},
                ],
            }
        ],
        data_row_expected=0.6,
        expected={0.0: 2, 1.0: 3},
    )


@pytest.fixture
def partially_matching_checklist_2():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                    {"schemaId": "ckppidq4u0002aeyjmcc4toxw"},
                    {"schemaId": "ckppiebx80004aeyjuwvos69e"},
                ],
            }
        ],
        data_row_expected=0.5,
        expected={1.0: 2, 0.0: 2},
    )


@pytest.fixture
def partially_matching_checklist_3():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                    {"schemaId": "ckppidq4u0002aeyjmcc4toxw"},
                    {"schemaId": "ckppiebx80004aeyjuwvos69e"},
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answers": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                    },
                    {"schemaId": "ckppide010001aeyj0yhiaghc"},
                ],
            }
        ],
        data_row_expected=0.5,
        expected={1.0: 2, 0.0: 2},
    )


@pytest.fixture
def empty_checklist_label():
    return NameSpace(
        labels=[],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answers": [{"schemaId": "ckppid25v0000aeyjmxfwlc7t"}],
            }
        ],
        data_row_expected=0.0,
        expected={0.0: 1},
    )


@pytest.fixture
def empty_checklist_prediction():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answers": [{"schemaId": "ckppid25v0000aeyjmxfwlc7t"}],
            }
        ],
        predictions=[],
        data_row_expected=0.0,
        expected={0.0: 1},
    )


@pytest.fixture
def matching_text():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "answer": "test",
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answer": "test",
            }
        ],
        expected=1.0,
    )


@pytest.fixture
def not_matching_text():
    return NameSpace(
        labels=[],
        classifications=[
            {
                "featureId": "1234567890111213141516171",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "answer": "test",
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "answer": "not_test",
            }
        ],
        expected=0.0,
    )


@pytest.fixture
def test_box_with_subclass():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "bbox": {
                    "top": 1099,
                    "left": 2010,
                    "height": 690,
                    "width": 591,
                },
                "classifications": [
                    {"schemaId": "ckppid25v0000aeyjmxfwlc7t", "answer": "test"}
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "bbox": {
                    "top": 1099,
                    "left": 2010,
                    "height": 690,
                    "width": 591,
                },
                "classifications": [
                    {"schemaId": "ckppid25v0000aeyjmxfwlc7t", "answer": "test"}
                ],
            }
        ],
        expected=1.0,
    )


@pytest.fixture
def test_box_with_wrong_subclass():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "bbox": {
                    "top": 1099,
                    "left": 2010,
                    "height": 690,
                    "width": 591,
                },
                "classifications": [
                    {"schemaId": "ckppid25v0000aeyjmxfwlc7t", "answer": "test"}
                ],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "bbox": {
                    "top": 1099,
                    "left": 2010,
                    "height": 690,
                    "width": 591,
                },
                "classifications": [
                    {
                        "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                        "answer": "not_test",
                    }
                ],
            }
        ],
        expected=0.5,
        expected_without_subclasses=1.0,
    )


@pytest.fixture
def line_pair():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "line": [{"x": 0, "y": 100}, {"x": 0, "y": 0}],
            }
        ],
        predictions=[
            {
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "line": [{"x": 5, "y": 95}, {"x": 0, "y": 0}],
            }
        ],
        expected=0.9496975567603978,
    )


@pytest.fixture
def point_pair():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "point": {"x": 0, "y": 0},
            }
        ],
        predictions=[
            {
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "point": {"x": 5, "y": 5},
            }
        ],
        expected=0.879113232477017,
    )


@pytest.fixture
def matching_ner():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "format": "text.location",
                "data": {"location": {"start": 0, "end": 10}},
            }
        ],
        predictions=[
            {
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "location": {"start": 0, "end": 10},
            }
        ],
        expected=1,
    )


@pytest.fixture
def no_matching_ner():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "format": "text.location",
                "data": {"location": {"start": 0, "end": 5}},
            }
        ],
        predictions=[
            {
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "location": {"start": 5, "end": 10},
            }
        ],
        expected=0,
    )


@pytest.fixture
def partial_matching_ner():
    return NameSpace(
        labels=[
            {
                "featureId": "ckppivl7p0006aeyj92cezr9d",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "format": "text.location",
                "data": {"location": {"start": 0, "end": 7}},
            }
        ],
        predictions=[
            {
                "dataRow": {"id": "ckppihxc10005aeyjen11h7jh"},
                "uuid": "76e0dcea-fe46-43e5-95f5-a5e3f378520a",
                "schemaId": "ckppid25v0000aeyjmxfwlc7t",
                "location": {"start": 3, "end": 5},
            }
        ],
        expected=0.2857142857142857,
    )



from enum import Enum, auto

from labelbox.orm.db_object import DbObject, Updateable, Deletable
from labelbox.orm.model import Field, Relationship


class Review(DbObject, Deletable, Updateable):
    """Reviewing labeled data is a collaborative quality assurance technique.

    A Review object indicates the quality of the assigned Label. The aggregated
    review numbers can be obtained on a Project object.

    Attributes:
        created_at (datetime)
        updated_at (datetime)
        score (float)

        created_by (Relationship): `ToOne` relationship to User
        organization (Relationship): `ToOne` relationship to Organization
        project (Relationship): `ToOne` relationship to Project
        label (Relationship): `ToOne` relationship to Label
    """

    class NetScore(Enum):
        """Negative, Zero, or Positive."""

        Negative = auto()
        Zero = auto()
        Positive = auto()

    updated_at = Field.DateTime("updated_at")
    created_at = Field.DateTime("created_at")
    score = Field.Float("score")

    created_by = Relationship.ToOne("User", False, "created_by")
    organization = Relationship.ToOne("Organization", False)
    project = Relationship.ToOne("Project", False)
    label = Relationship.ToOne("Label", False)



from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
from labelbox.data.serialization.ndjson.converter import NDJsonConverter
from labelbox.types import Label, ClassificationAnnotation, Text


def test_generic_data_row_global_key():
    label_1 = Label(
        data=GenericDataRowData(global_key="test"),
        annotations=[
            ClassificationAnnotation(
                name="free_text",
                value=Text(answer="sample text"),
                extra={"uuid": "141c3592-e5f0-4866-9943-d4a21fd47eb0"},
            )
        ],
    )
    label_2 = Label(
        data={"global_key": "test"},
        annotations=[
            ClassificationAnnotation(
                name="free_text",
                value=Text(answer="sample text"),
                extra={"uuid": "141c3592-e5f0-4866-9943-d4a21fd47eb0"},
            )
        ],
    )

    expected_result = [
        {
            "answer": "sample text",
            "dataRow": {"globalKey": "test"},
            "name": "free_text",
            "uuid": "141c3592-e5f0-4866-9943-d4a21fd47eb0",
        }
    ]
    assert (
        list(NDJsonConverter.serialize([label_1]))
        == list(NDJsonConverter.serialize([label_2]))
        == expected_result
    )


def test_generic_data_row_id():
    label_1 = Label(
        data=GenericDataRowData(uid="test"),
        annotations=[
            ClassificationAnnotation(
                name="free_text",
                value=Text(answer="sample text"),
                extra={"uuid": "141c3592-e5f0-4866-9943-d4a21fd47eb0"},
            )
        ],
    )
    label_2 = Label(
        data={"uid": "test"},
        annotations=[
            ClassificationAnnotation(
                name="free_text",
                value=Text(answer="sample text"),
                extra={"uuid": "141c3592-e5f0-4866-9943-d4a21fd47eb0"},
            )
        ],
    )

    expected_result = [
        {
            "answer": "sample text",
            "dataRow": {"id": "test"},
            "name": "free_text",
            "uuid": "141c3592-e5f0-4866-9943-d4a21fd47eb0",
        }
    ]
    assert (
        list(NDJsonConverter.serialize([label_1]))
        == list(NDJsonConverter.serialize([label_2]))
        == expected_result
    )



from typing import List, Optional, Tuple, Union
from itertools import product

import numpy as np
from shapely.geometry import Polygon

from ..group import (
    get_feature_pairs,
    get_identifying_key,
    has_no_annotations,
    has_no_matching_annotations,
)
from ...annotation_types import (
    ObjectAnnotation,
    ClassificationAnnotation,
    Mask,
    Geometry,
    Point,
    Line,
    Checklist,
    Text,
    TextEntity,
    Radio,
    ScalarMetricValue,
)


def miou(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    include_subclasses: bool,
) -> Optional[ScalarMetricValue]:
    """
    Computes miou for an arbitrary set of ground truth and predicted annotations.
    It first computes the iou for each metric and then takes the average (weighting each class equally)

    Args:
        ground_truth : Label containing human annotations or annotations known to be correct
        prediction: Label representing model predictions
        include_subclasses (bool): Whether or not to include subclasses in the iou calculation.
            If set to True, the iou between two overlapping objects of the same type is 0 if the subclasses are not the same.
    Returns:
        float indicating the iou score for all features represented in the annotations passed to this function.
        Returns None if there are no annotations in ground_truth or prediction annotations
    """
    annotation_pairs = get_feature_pairs(predictions, ground_truths)
    ious = [
        feature_miou(annotation_pair[0], annotation_pair[1], include_subclasses)
        for annotation_pair in annotation_pairs.values()
    ]
    ious = [iou for iou in ious if iou is not None]
    return None if not len(ious) else np.mean(ious)


def feature_miou(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    include_subclasses: bool,
) -> Optional[ScalarMetricValue]:
    """
    Computes iou score for all features of the same class.

    Args:
        ground_truths: List of ground truth annotations with the same feature schema.
        predictions: List of annotations with the same feature schema.
        include_subclasses (bool): Whether or not to include subclasses in the iou calculation.
            If set to True, the iou between two overlapping objects of the same type is 0 if the subclasses are not the same.
    Returns:
        float representing the iou score for the feature type if score can be computed otherwise None.
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return 0.0
    elif has_no_annotations(ground_truths, predictions):
        return None
    elif isinstance(predictions[0].value, Mask):
        return mask_miou(ground_truths, predictions, include_subclasses)
    elif isinstance(predictions[0].value, Geometry):
        return vector_miou(ground_truths, predictions, include_subclasses)
    elif isinstance(predictions[0], ClassificationAnnotation):
        return classification_miou(ground_truths, predictions)
    elif isinstance(predictions[0].value, TextEntity):
        return ner_miou(ground_truths, predictions, include_subclasses)
    else:
        raise ValueError(
            f"Unexpected annotation found. Found {type(predictions[0].value)}"
        )


def vector_miou(
    ground_truths: List[ObjectAnnotation],
    predictions: List[ObjectAnnotation],
    include_subclasses: bool,
    buffer=70.0,
) -> Optional[ScalarMetricValue]:
    """
    Computes iou score for all features with the same feature schema id.
    Calculation includes subclassifications.

    Args:
        ground_truths: List of ground truth vector annotations
        predictions: List of prediction vector annotations
    Returns:
        float representing the iou score for the feature type.
         If there are no matches then this returns none
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return 0.0
    elif has_no_annotations(ground_truths, predictions):
        return None
    pairs = _get_vector_pairs(ground_truths, predictions, buffer=buffer)
    return object_pair_miou(pairs, include_subclasses)


def object_pair_miou(
    pairs: List[Tuple[ObjectAnnotation, ObjectAnnotation, ScalarMetricValue]],
    include_subclasses,
) -> ScalarMetricValue:
    pairs.sort(key=lambda triplet: triplet[2], reverse=True)
    solution_agreements = []
    solution_features = set()
    all_features = set()
    for prediction, ground_truth, agreement in pairs:
        all_features.update({id(prediction), id(ground_truth)})
        if (
            id(prediction) not in solution_features
            and id(ground_truth) not in solution_features
        ):
            solution_features.update({id(prediction), id(ground_truth)})
            if include_subclasses:
                classification_iou = miou(
                    prediction.classifications,
                    ground_truth.classifications,
                    include_subclasses=False,
                )
                classification_iou = (
                    classification_iou
                    if classification_iou is not None
                    else agreement
                )
                solution_agreements.append(
                    (agreement + classification_iou) / 2.0
                )
            else:
                solution_agreements.append(agreement)

    # Add zeros for unmatched Features
    solution_agreements.extend(
        [0.0] * (len(all_features) - len(solution_features))
    )
    return np.mean(solution_agreements)


def mask_miou(
    ground_truths: List[ObjectAnnotation],
    predictions: List[ObjectAnnotation],
    include_subclasses: bool,
) -> Optional[ScalarMetricValue]:
    """
    Computes iou score for all features with the same feature schema id.
    Calculation includes subclassifications.

    Args:
        ground_truths: List of ground truth mask annotations
        predictions: List of prediction mask annotations
    Returns:
        float representing the iou score for the masks
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return 0.0
    elif has_no_annotations(ground_truths, predictions):
        return None

    if include_subclasses:
        pairs = _get_mask_pairs(ground_truths, predictions)
        return object_pair_miou(pairs, include_subclasses=include_subclasses)

    prediction_np = np.max(
        [pred.value.draw(color=1) for pred in predictions], axis=0
    )
    ground_truth_np = np.max(
        [ground_truth.value.draw(color=1) for ground_truth in ground_truths],
        axis=0,
    )
    if prediction_np.shape != ground_truth_np.shape:
        raise ValueError(
            "Prediction and mask must have the same shape."
            f" Found {prediction_np.shape}/{ground_truth_np.shape}."
        )

    return _mask_iou(ground_truth_np, prediction_np)


def classification_miou(
    ground_truths: List[ClassificationAnnotation],
    predictions: List[ClassificationAnnotation],
) -> ScalarMetricValue:
    """
    Computes iou score for all features with the same feature schema id.

    Args:
        ground_truths: List of ground truth classification annotations
        predictions: List of prediction classification annotations
    Returns:
        float representing the iou score for the classification
    """

    if len(predictions) != len(ground_truths) != 1:
        return 0.0

    prediction, ground_truth = predictions[0], ground_truths[0]

    if type(prediction) is not type(ground_truth):
        raise TypeError(
            "Classification features must be the same type to compute agreement. "
            f"Found `{type(prediction)}` and `{type(ground_truth)}`"
        )

    if isinstance(prediction.value, Text):
        return text_iou(ground_truth.value, prediction.value)
    elif isinstance(prediction.value, Radio):
        return radio_iou(ground_truth.value, prediction.value)
    elif isinstance(prediction.value, Checklist):
        return checklist_iou(ground_truth.value, prediction.value)
    else:
        raise ValueError(f"Unsupported subclass. {prediction}.")


def radio_iou(ground_truth: Radio, prediction: Radio) -> ScalarMetricValue:
    """
    Calculates agreement between ground truth and predicted radio values
    """
    key = get_identifying_key([prediction.answer], [ground_truth.answer])
    return float(
        getattr(prediction.answer, key) == getattr(ground_truth.answer, key)
    )


def text_iou(ground_truth: Text, prediction: Text) -> ScalarMetricValue:
    """
    Calculates agreement between ground truth and predicted text
    """
    return float(prediction.answer == ground_truth.answer)


def checklist_iou(
    ground_truth: Checklist, prediction: Checklist
) -> ScalarMetricValue:
    """
    Calculates agreement between ground truth and predicted checklist items
    """
    key = get_identifying_key(prediction.answer, ground_truth.answer)
    schema_ids_pred = {getattr(answer, key) for answer in prediction.answer}
    schema_ids_label = {getattr(answer, key) for answer in ground_truth.answer}
    return float(
        len(schema_ids_label & schema_ids_pred)
        / len(schema_ids_label | schema_ids_pred)
    )


def _get_vector_pairs(
    ground_truths: List[ObjectAnnotation],
    predictions: List[ObjectAnnotation],
    buffer: float,
) -> List[Tuple[ObjectAnnotation, ObjectAnnotation, ScalarMetricValue]]:
    """
    # Get iou score for all pairs of ground truths and predictions
    """
    pairs = []
    for ground_truth, prediction in product(ground_truths, predictions):
        if isinstance(prediction.value, Geometry) and isinstance(
            ground_truth.value, Geometry
        ):
            if isinstance(prediction.value, (Line, Point)):
                score = _polygon_iou(
                    prediction.value.shapely.buffer(buffer),
                    ground_truth.value.shapely.buffer(buffer),
                )
            else:
                score = _polygon_iou(
                    prediction.value.shapely, ground_truth.value.shapely
                )
            pairs.append((ground_truth, prediction, score))
    return pairs


def _get_mask_pairs(
    ground_truths: List[ObjectAnnotation], predictions: List[ObjectAnnotation]
) -> List[Tuple[ObjectAnnotation, ObjectAnnotation, ScalarMetricValue]]:
    """
    # Get iou score for all pairs of ground truths and predictions
    """
    pairs = []
    for ground_truth, prediction in product(ground_truths, predictions):
        if isinstance(prediction.value, Mask) and isinstance(
            ground_truth.value, Mask
        ):
            score = _mask_iou(
                prediction.value.draw(color=1), ground_truth.value.draw(color=1)
            )
            pairs.append((ground_truth, prediction, score))
    return pairs


def _polygon_iou(poly1: Polygon, poly2: Polygon) -> ScalarMetricValue:
    """Computes iou between two shapely polygons."""
    poly1, poly2 = _ensure_valid_poly(poly1), _ensure_valid_poly(poly2)
    if poly1.intersects(poly2):
        return poly1.intersection(poly2).area / poly1.union(poly2).area
    return 0.0


def _ensure_valid_poly(poly):
    if not poly.is_valid:
        return poly.buffer(0)
    return poly


def _mask_iou(mask1: np.ndarray, mask2: np.ndarray) -> ScalarMetricValue:
    """Computes iou between two binary segmentation masks."""
    return np.sum(mask1 & mask2) / np.sum(mask1 | mask2)


def _get_ner_pairs(
    ground_truths: List[ObjectAnnotation], predictions: List[ObjectAnnotation]
) -> List[Tuple[ObjectAnnotation, ObjectAnnotation, ScalarMetricValue]]:
    """Get iou score for all possible pairs of ground truths and predictions"""
    pairs = []
    for ground_truth, prediction in product(ground_truths, predictions):
        score = _ner_iou(ground_truth.value, prediction.value)
        pairs.append((ground_truth, prediction, score))
    return pairs


def _ner_iou(ner1: TextEntity, ner2: TextEntity):
    """Computes iou between two text entity annotations"""
    intersection_start, intersection_end = (
        max(ner1.start, ner2.start),
        min(ner1.end, ner2.end),
    )
    union_start, union_end = (
        min(ner1.start, ner2.start),
        max(ner1.end, ner2.end),
    )
    # edge case of only one character in text
    if union_start == union_end:
        return 1
    # if there is no intersection
    if intersection_start > intersection_end:
        return 0
    return (intersection_end - intersection_start) / (union_end - union_start)


def ner_miou(
    ground_truths: List[ObjectAnnotation],
    predictions: List[ObjectAnnotation],
    include_subclasses: bool,
) -> Optional[ScalarMetricValue]:
    """
    Computes iou score for all features with the same feature schema id.
    Calculation includes subclassifications.

    Args:
        ground_truths: List of ground truth ner annotations
        predictions: List of prediction ner annotations
    Returns:
        float representing the iou score for the feature type.
         If there are no matches then this returns none
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return 0.0
    elif has_no_annotations(ground_truths, predictions):
        return None
    pairs = _get_ner_pairs(ground_truths, predictions)
    return object_pair_miou(pairs, include_subclasses)



from labelbox.orm.db_object import DbObject
from labelbox.orm.model import Field


class ModelConfig(DbObject):
    """A ModelConfig represents a set of inference params configured for a model

    Attributes:
        inference_params (JSON): Dict of inference params
        model_id (str): ID of the model to configure
        name (str): Name of config
    """

    inference_params = Field.Json("inference_params", "inferenceParams")
    model_id = Field.String("model_id", "modelId")
    name = Field.String("name", "name")



from collections import defaultdict
import copy
from itertools import groupby
from operator import itemgetter
from typing import Generator, List, Tuple, Union
from uuid import uuid4

from pydantic import BaseModel

from ...annotation_types.annotation import (
    ClassificationAnnotation,
    ObjectAnnotation,
)
from ...annotation_types.collection import LabelCollection
from ...annotation_types.label import Label
from ...annotation_types.llm_prompt_response.prompt import (
    PromptClassificationAnnotation,
)
from ...annotation_types.metrics import ConfusionMatrixMetric, ScalarMetric
from ...annotation_types.mmc import MessageEvaluationTaskAnnotation
from ...annotation_types.relationship import RelationshipAnnotation
from ...annotation_types.video import (
    VideoClassificationAnnotation,
    VideoMaskAnnotation,
    VideoObjectAnnotation,
)
from .classification import (
    NDChecklistSubclass,
    NDClassification,
    NDClassificationType,
    NDPromptClassification,
    NDPromptClassificationType,
    NDPromptText,
    NDRadioSubclass,
)
from .metric import NDConfusionMatrixMetric, NDMetricAnnotation, NDScalarMetric
from .mmc import NDMessageTask
from .objects import (
    NDObject,
    NDObjectType,
    NDSegments,
    NDVideoMasks,
)
from .relationship import NDRelationship

AnnotationType = Union[
    NDObjectType,
    NDClassificationType,
    NDPromptClassificationType,
    NDConfusionMatrixMetric,
    NDScalarMetric,
    NDSegments,
    NDVideoMasks,
    NDRelationship,
    NDPromptText,
    NDMessageTask,
]


class NDLabel(BaseModel):
    annotations: AnnotationType

    @classmethod
    def from_common(
        cls, data: LabelCollection
    ) -> Generator["NDLabel", None, None]:
        for label in data:
            yield from cls._create_relationship_annotations(label)
            yield from cls._create_non_video_annotations(label)
            yield from cls._create_video_annotations(label)

    @staticmethod
    def _get_consecutive_frames(
        frames_indices: List[int],
    ) -> List[Tuple[int, int]]:
        consecutive = []
        for k, g in groupby(enumerate(frames_indices), lambda x: x[0] - x[1]):
            group = list(map(itemgetter(1), g))
            consecutive.append((group[0], group[-1]))
        return consecutive

    @classmethod
    def _get_segment_frame_ranges(
        cls,
        annotation_group: List[
            Union[VideoClassificationAnnotation, VideoObjectAnnotation]
        ],
    ) -> List[Tuple[int, int]]:
        sorted_frame_segment_indices = sorted(
            [
                (annotation.frame, annotation.segment_index)
                for annotation in annotation_group
                if annotation.segment_index is not None
            ]
        )
        if len(sorted_frame_segment_indices) == 0:
            # Group segment by consecutive frames, since `segment_index` is not present
            return cls._get_consecutive_frames(
                sorted([annotation.frame for annotation in annotation_group])
            )
        elif len(sorted_frame_segment_indices) == len(annotation_group):
            # Group segment by segment_index
            last_segment_id = 0
            segment_groups = defaultdict(list)
            for frame, segment_index in sorted_frame_segment_indices:
                if segment_index < last_segment_id:
                    raise ValueError(
                        f"`segment_index` must be in ascending order. Please investigate video annotation at frame, '{frame}'"
                    )
                segment_groups[segment_index].append(frame)
                last_segment_id = segment_index
            frame_ranges = []
            for group in segment_groups.values():
                frame_ranges.append((group[0], group[-1]))
            return frame_ranges
        else:
            raise ValueError(
                f"Video annotations cannot partially have `segment_index` set"
            )

    @classmethod
    def _create_video_annotations(
        cls, label: Label
    ) -> Generator[Union[NDChecklistSubclass, NDRadioSubclass], None, None]:
        video_annotations = defaultdict(list)
        for annot in label.annotations:
            if isinstance(
                annot, (VideoClassificationAnnotation, VideoObjectAnnotation)
            ):
                video_annotations[annot.feature_schema_id or annot.name].append(
                    annot
                )
            elif isinstance(annot, VideoMaskAnnotation):
                yield NDObject.from_common(annotation=annot, data=label.data)

        for annotation_group in video_annotations.values():
            segment_frame_ranges = cls._get_segment_frame_ranges(
                annotation_group
            )
            if isinstance(annotation_group[0], VideoClassificationAnnotation):
                annotation = annotation_group[0]
                frames_data = []
                for frames in segment_frame_ranges:
                    frames_data.append({"start": frames[0], "end": frames[-1]})
                annotation.extra.update({"frames": frames_data})
                yield NDClassification.from_common(annotation, label.data)

            elif isinstance(annotation_group[0], VideoObjectAnnotation):
                segments = []
                for start_frame, end_frame in segment_frame_ranges:
                    segment = []
                    for annotation in annotation_group:
                        if (
                            annotation.keyframe
                            and start_frame <= annotation.frame <= end_frame
                        ):
                            segment.append(annotation)
                    segments.append(segment)
                yield NDObject.from_common(segments, label.data)

    @classmethod
    def _create_non_video_annotations(cls, label: Label):
        non_video_annotations = [
            annot
            for annot in label.annotations
            if not isinstance(
                annot,
                (
                    VideoClassificationAnnotation,
                    VideoObjectAnnotation,
                    VideoMaskAnnotation,
                ),
            )
        ]
        for annotation in non_video_annotations:
            if isinstance(annotation, ClassificationAnnotation):
                yield NDClassification.from_common(annotation, label.data)
            elif isinstance(annotation, ObjectAnnotation):
                yield NDObject.from_common(annotation, label.data)
            elif isinstance(annotation, (ScalarMetric, ConfusionMatrixMetric)):
                yield NDMetricAnnotation.from_common(annotation, label.data)
            elif isinstance(annotation, RelationshipAnnotation):
                yield NDRelationship.from_common(annotation, label.data)
            elif isinstance(annotation, PromptClassificationAnnotation):
                yield NDPromptClassification.from_common(annotation, label.data)
            elif isinstance(annotation, MessageEvaluationTaskAnnotation):
                yield NDMessageTask.from_common(annotation, label.data)
            else:
                raise TypeError(
                    f"Unable to convert object to MAL format. `{type(getattr(annotation, 'value',annotation))}`"
                )

    @classmethod
    def _create_relationship_annotations(cls, label: Label):
        for annotation in label.annotations:
            if isinstance(annotation, RelationshipAnnotation):
                uuid1 = uuid4()
                uuid2 = uuid4()
                source = copy.copy(annotation.value.source)
                target = copy.copy(annotation.value.target)
                if not isinstance(source, ObjectAnnotation) or not isinstance(
                    target, ObjectAnnotation
                ):
                    raise TypeError(
                        f"Unable to create relationship with non ObjectAnnotations. `Source: {type(source)} Target: {type(target)}`"
                    )
                if not source._uuid:
                    source._uuid = uuid1
                if not target._uuid:
                    target._uuid = uuid2
                yield NDRelationship.from_common(annotation, label.data)



from labelbox.orm.db_object import DbObject, Updateable
from labelbox.orm.model import Field


class ProjectResourceTag(DbObject, Updateable):
    """Project resource tag to associate ProjectResourceTag to Project.

    Attributes:
        resourceTagId (str)
        projectId (str)

        resource_tag (Relationship): `ToOne` relationship to ResourceTag
    """

    resource_tag_id = Field.ID("resource_tag_id")
    project_id = Field.ID("project_id")



from datetime import datetime, timedelta, timezone


def test_dates(project):
    assert isinstance(project.created_at, datetime)
    assert isinstance(project.updated_at, datetime)

    project.update(setup_complete=datetime.now())
    assert isinstance(project.setup_complete, datetime)


def test_utc_conversion(project):
    assert isinstance(project.created_at, datetime)
    assert project.created_at.tzinfo == timezone.utc

    # Update with a datetime without TZ info.
    project.update(setup_complete=datetime.now())
    # Check that the server-side, UTC date is the same as local date
    # converted locally to UTC.
    diff = project.setup_complete - datetime.now().astimezone(timezone.utc)
    assert abs(diff) < timedelta(minutes=1)

    # Update with a datetime with TZ info
    tz = timezone(timedelta(hours=6))  # +6 timezone
    project.update(setup_complete=datetime.now(timezone.utc).replace(tzinfo=tz))
    diff = datetime.now(timezone.utc) - project.setup_complete.replace(
        tzinfo=timezone.utc
    )
    assert diff > timedelta(hours=5, minutes=58)



from typing import List

from labelbox.schema.internal.data_row_upsert_item import (
    DataRowItemBase,
)
from labelbox.schema.internal.descriptor_file_creator import (
    DescriptorFileCreator,
)
from pydantic import BaseModel


class UploadManifest(BaseModel):
    source: str
    item_count: int
    chunk_uris: List[str]


SOURCE_SDK = "SDK"


def upload_in_chunks(
    client,
    specs: List[DataRowItemBase],
    file_upload_thread_count: int,
    max_chunk_size_bytes: int,
) -> UploadManifest:
    empty_specs = list(filter(lambda spec: spec.is_empty(), specs))
    if empty_specs:
        ids = list(map(lambda spec: spec.id.get("value"), empty_specs))
        ids = list(filter(lambda x: x is not None and len(x) > 0, ids))
        if len(ids) > 0:
            raise ValueError(
                f"The following items have an empty payload: {ids}"
            )
        else:  # case of create items
            raise ValueError("Some items have an empty payload")

    chunk_uris = DescriptorFileCreator(client).create(
        specs, max_chunk_size_bytes=max_chunk_size_bytes
    )

    return UploadManifest(
        source=SOURCE_SDK, item_count=len(specs), chunk_uris=chunk_uris
    )



from typing import List
from uuid import uuid4

import pytest
from lbox.exceptions import (
    LabelboxError,
    MalformedQueryException,
    ProcessingWaitTimeout,
    ResourceConflict,
)

from labelbox import Dataset, Project


def get_data_row_ids(ds: Dataset):
    return [dr.uid for dr in list(ds.data_rows())]


def test_create_batch(project: Project, big_dataset_data_row_ids: List[str]):
    batch = project.create_batch(
        "test-batch",
        big_dataset_data_row_ids,
        3,
        consensus_settings={"number_of_labels": 3, "coverage_percentage": 0.1},
    )

    assert batch.name == "test-batch"
    assert batch.size == len(big_dataset_data_row_ids)
    assert len([dr for dr in batch.failed_data_row_ids]) == 0


def test_create_batch_with_invalid_data_rows_ids(project: Project):
    with pytest.raises(MalformedQueryException) as ex:
        project.create_batch("test-batch", data_rows=["a", "b", "c"])
        assert (
            str(ex) == "No valid data rows to be added from the list provided!"
        )


def test_create_batch_with_the_same_name(
    project: Project, small_dataset: Dataset
):
    batch1 = project.create_batch(
        "batch1", data_rows=get_data_row_ids(small_dataset)
    )
    assert batch1.name == "batch1"

    with pytest.raises(ResourceConflict):
        project.create_batch(
            "batch1", data_rows=get_data_row_ids(small_dataset)
        )


def test_create_batch_with_same_data_row_ids(
    project: Project, small_dataset: Dataset
):
    batch1 = project.create_batch(
        "batch1", data_rows=get_data_row_ids(small_dataset)
    )
    assert batch1.name == "batch1"

    with pytest.raises(MalformedQueryException) as ex:
        project.create_batch(
            "batch2", data_rows=get_data_row_ids(small_dataset)
        )
        assert str(ex) == "No valid data rows to add to project"


def test_create_batch_with_non_existent_global_keys(project: Project):
    with pytest.raises(MalformedQueryException) as ex:
        project.create_batch("batch1", global_keys=["key1"])
        assert (
            str(ex)
            == "Data rows with the following global keys do not exist: key1."
        )


def test_create_batch_with_string_priority(
    project: Project, small_dataset: Dataset
):
    with pytest.raises(LabelboxError):
        project.create_batch(
            "batch1", data_rows=get_data_row_ids(small_dataset), priority="abcd"
        )


def test_create_batch_with_null_priority(
    project: Project, small_dataset: Dataset
):
    with pytest.raises(LabelboxError):
        project.create_batch(
            "batch1", data_rows=get_data_row_ids(small_dataset), priority=None
        )


def test_create_batch_async(
    project: Project, big_dataset_data_row_ids: List[str]
):
    batch = project._create_batch_async(
        "big-batch", big_dataset_data_row_ids, priority=3
    )
    assert batch.name == "big-batch"
    assert batch.size == len(big_dataset_data_row_ids)
    assert len([dr for dr in batch.failed_data_row_ids]) == 0


def test_create_batch_with_consensus_settings(
    project: Project, small_dataset: Dataset
):
    export_task = small_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]
    consensus_settings = {"coverage_percentage": 0.1, "number_of_labels": 3}
    batch = project.create_batch(
        "batch with consensus settings",
        data_rows,
        3,
        consensus_settings=consensus_settings,
    )
    assert batch.name == "batch with consensus settings"
    assert batch.size == len(data_rows)
    assert batch.consensus_settings == consensus_settings


def test_create_batch_with_data_row_class(
    project: Project, small_dataset: Dataset
):
    export_task = small_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]
    batch = project.create_batch("test-batch-data-rows", data_rows, 3)
    assert batch.name == "test-batch-data-rows"
    assert batch.size == len(data_rows)


def test_archive_batch(project: Project, small_dataset: Dataset):
    export_task = small_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]

    batch = project.create_batch("batch to archive", data_rows)
    batch.remove_queued_data_rows()
    overview = project.get_overview()

    assert overview.to_label == 0


def test_delete(project: Project, small_dataset: Dataset):
    export_task = small_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]
    batch = project.create_batch("batch to delete", data_rows)
    batch.delete()

    assert len(list(project.batches())) == 0


def test_batch_project(project: Project, small_dataset: Dataset):
    export_task = small_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]
    batch = project.create_batch(
        "batch to test project relationship", data_rows
    )

    project_from_batch = batch.project()

    assert project_from_batch.uid == project.uid
    assert project_from_batch.name == project.name


def test_batch_creation_for_data_rows_with_issues(
    project: Project,
    small_dataset: Dataset,
    dataset_with_invalid_data_rows: Dataset,
):
    """
    Create a batch containing both valid and invalid data rows
    """
    valid_data_rows = [dr.uid for dr in list(small_dataset.data_rows())]
    invalid_data_rows = [
        dr.uid for dr in list(dataset_with_invalid_data_rows.data_rows())
    ]
    data_rows_to_add = valid_data_rows + invalid_data_rows

    assert len(data_rows_to_add) == 4
    batch = project.create_batch(
        "batch to test failed data rows", data_rows_to_add
    )
    failed_data_row_ids = [x for x in batch.failed_data_row_ids]
    assert len(failed_data_row_ids) == 2

    failed_data_row_ids_set = set(failed_data_row_ids)
    invalid_data_rows_set = set(invalid_data_rows)
    assert len(failed_data_row_ids_set.intersection(invalid_data_rows_set)) == 2


def test_batch_creation_with_processing_timeout(
    project: Project,
    small_dataset: Dataset,
    unique_dataset: Dataset,
    upload_invalid_data_rows_for_dataset,
):
    """
    Create a batch with zero wait time, this means that the waiting logic will throw exception immediately
    """
    #  wait for these data rows to be processed
    valid_data_rows = [dr.uid for dr in list(small_dataset.data_rows())]

    # upload data rows for this dataset and don't wait
    upload_invalid_data_rows_for_dataset(unique_dataset)
    unprocessed_data_rows = [dr.uid for dr in list(unique_dataset.data_rows())]

    data_row_ids = valid_data_rows + unprocessed_data_rows

    stashed_wait_timeout = project._wait_processing_max_seconds
    with pytest.raises(ProcessingWaitTimeout):
        # emulate the situation where there are still some data rows being
        # processed but wait timeout exceeded
        project._wait_processing_max_seconds = 0
        project.create_batch("batch to test failed data rows", data_row_ids)
    project._wait_processing_max_seconds = stashed_wait_timeout


def test_list_all_batches(project: Project, client, image_url: str):
    """
    Test to verify that we can retrieve all available batches in the project.
    """
    # Data to use
    img_assets = [
        {"row_data": image_url, "external_id": str(uuid4())}
        for asset in range(0, 2)
    ]
    data = [img_assets for _ in range(0, 2)]

    # Setup
    batches = []
    datasets = []

    for assets in data:
        dataset = client.create_dataset(name=str(uuid4()))
        create_data_rows_task = dataset.create_data_rows(assets)
        create_data_rows_task.wait_till_done()
        datasets.append(dataset)

    for dataset in datasets:
        data_row_ids = get_data_row_ids(dataset)
        new_batch = project.create_batch(
            name=str(uuid4()), data_rows=data_row_ids
        )
        batches.append(new_batch)

    # Test
    project_batches = list(project.batches())
    assert len(batches) == len(project_batches)

    for project_batch in project_batches:
        for assets in data:
            assert len(assets) == project_batch.size

    # Clean up
    for dataset in datasets:
        dataset.delete()


def test_list_project_batches_with_no_batches(project: Project):
    batches = list(project.batches())
    assert len(batches) == 0


@pytest.mark.skip(
    reason="Test cannot be used effectively with MAL/LabelImport. \
Fix/Unskip after resolving deletion with MAL/LabelImport"
)
def test_delete_labels(project, small_dataset):
    export_task = small_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]
    batch = project.create_batch("batch to delete labels", data_rows)


@pytest.mark.skip(
    reason="Test cannot be used effectively with MAL/LabelImport. \
Fix/Unskip after resolving deletion with MAL/LabelImport"
)
def test_delete_labels_with_templates(project: Project, small_dataset: Dataset):
    export_task = small_dataset.export()
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    data_rows = [dr.json["data_row"]["id"] for dr in stream]
    batch = project.create_batch(
        "batch to delete labels w templates", data_rows
    )

    export_task = project.export(filters={"batch_ids": [batch.uid]})
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    exported_data_rows = [dr.json["data_row"]["id"] for dr in stream]

    res = batch.delete_labels(labels_as_template=True)

    export_task = project.export(filters={"batch_ids": [batch.uid]})
    export_task.wait_till_done()
    stream = export_task.get_buffered_stream()
    exported_data_rows = [dr.json["data_row"]["id"] for dr in stream]

    assert len(exported_data_rows) == 5



from typing import List, Optional, Tuple, Union

import numpy as np

from ..iou.calculation import (
    _get_mask_pairs,
    _get_vector_pairs,
    _get_ner_pairs,
    miou,
)
from ...annotation_types import (
    ObjectAnnotation,
    ClassificationAnnotation,
    Mask,
    Geometry,
    Checklist,
    Radio,
    TextEntity,
    ScalarMetricValue,
    ConfusionMatrixMetricValue,
)
from ..group import (
    get_feature_pairs,
    get_identifying_key,
    has_no_annotations,
    has_no_matching_annotations,
)


def confusion_matrix(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    include_subclasses: bool,
    iou: float,
) -> ConfusionMatrixMetricValue:
    """
    Computes the confusion matrix for an arbitrary set of ground truth and predicted annotations.
    It first computes the confusion matrix for each metric and then sums across all classes

    Args:
        ground_truth : Label containing human annotations or annotations known to be correct
        prediction: Label representing model predictions
        include_subclasses (bool): Whether or not to include subclasses in the calculation.
            If set to True, the iou between two overlapping objects of the same type is 0 if the subclasses are not the same.
        iou: minimum overlap between objects for them to count as matching
    Returns:
        confusion matrix as a list: [TP,FP,TN,FN]
        Returns None if there are no annotations in ground_truth or prediction annotations
    """

    annotation_pairs = get_feature_pairs(ground_truths, predictions)
    conf_matrix = [
        feature_confusion_matrix(
            annotation_pair[0], annotation_pair[1], include_subclasses, iou
        )
        for annotation_pair in annotation_pairs.values()
    ]
    matrices = [matrix for matrix in conf_matrix if matrix is not None]
    return None if not len(matrices) else np.sum(matrices, axis=0).tolist()


def feature_confusion_matrix(
    ground_truths: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    predictions: List[Union[ObjectAnnotation, ClassificationAnnotation]],
    include_subclasses: bool,
    iou: float,
) -> Optional[ConfusionMatrixMetricValue]:
    """
    Computes confusion matrix for all features of the same class.

    Args:
        ground_truths: List of ground truth annotations belonging to the same class.
        predictions: List of annotations  belonging to the same class.
        include_subclasses (bool): Whether or not to include subclasses in the calculation.
            If set to True, the iou between two overlapping objects of the same type is 0 if the subclasses are not the same.
    Returns:
        confusion matrix as a list: [TP,FP,TN,FN]
        Returns None if there are no annotations in ground_truth or prediction annotations
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return [0, len(predictions), 0, len(ground_truths)]
    elif has_no_annotations(ground_truths, predictions):
        return None
    elif isinstance(predictions[0].value, Mask):
        return mask_confusion_matrix(
            ground_truths, predictions, include_subclasses, iou
        )
    elif isinstance(predictions[0].value, Geometry):
        return vector_confusion_matrix(
            ground_truths, predictions, include_subclasses, iou
        )
    elif isinstance(predictions[0].value, TextEntity):
        return ner_confusion_matrix(
            ground_truths, predictions, include_subclasses, iou
        )
    elif isinstance(predictions[0], ClassificationAnnotation):
        return classification_confusion_matrix(ground_truths, predictions)
    else:
        raise ValueError(
            f"Unexpected annotation found. Found {type(predictions[0].value)}"
        )


def classification_confusion_matrix(
    ground_truths: List[ClassificationAnnotation],
    predictions: List[ClassificationAnnotation],
) -> ConfusionMatrixMetricValue:
    """
    Computes iou score for all features with the same feature schema id.
    Because these predictions and ground truths are already sorted by schema id,
    there can only be one of each (or zero if the classification was not predicted or labeled).

    Args:
        ground_truths: List of ground truth classification annotations
        predictions: List of prediction classification annotations
    Returns:
        confusion matrix as a list: [TP,FP,TN,FN]
        Returns None if there are no annotations in ground_truth or prediction annotations
    """

    if has_no_matching_annotations(ground_truths, predictions):
        return [0, len(predictions), 0, len(ground_truths)]
    elif (
        has_no_annotations(ground_truths, predictions)
        or len(predictions) > 1
        or len(ground_truths) > 1
    ):
        # Note that we could return [0,0,0,0] but that will bloat the imports for no reason
        return None

    prediction, ground_truth = predictions[0], ground_truths[0]

    if type(prediction) is not type(ground_truth):
        raise TypeError(
            "Classification features must be the same type to compute agreement. "
            f"Found `{type(prediction)}` and `{type(ground_truth)}`"
        )

    if isinstance(prediction.value, Radio):
        return radio_confusion_matrix(ground_truth.value, prediction.value)
    elif isinstance(prediction.value, Checklist):
        return checklist_confusion_matrix(ground_truth.value, prediction.value)
    else:
        raise ValueError(
            f"Unsupported subclass. {prediction}. Only Radio and Checklist are supported"
        )


def vector_confusion_matrix(
    ground_truths: List[ObjectAnnotation],
    predictions: List[ObjectAnnotation],
    include_subclasses: bool,
    iou: float,
    buffer=70.0,
) -> Optional[ConfusionMatrixMetricValue]:
    """
    Computes confusion matrix for any vector class (point, polygon, line, rectangle).
    Ground truths and predictions should all belong to the same class.

    Args:
        ground_truths: List of ground truth vector annotations
        predictions: List of prediction vector annotations
        iou: minimum overlap between objects for them to count as matching
        include_subclasses (bool): Whether or not to include subclasses in the calculation.
            If set to True, the iou between two overlapping objects of the same type is 0 if the subclasses are not the same.
        buffer: How much to buffer point and lines (used for determining if overlap meets iou threshold )
    Returns:
        confusion matrix as a list: [TP,FP,TN,FN]
         Returns None if there are no annotations in ground_truth or prediction annotations
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return [0, len(predictions), 0, len(ground_truths)]
    elif has_no_annotations(ground_truths, predictions):
        return None

    pairs = _get_vector_pairs(ground_truths, predictions, buffer=buffer)
    return object_pair_confusion_matrix(pairs, include_subclasses, iou)


def object_pair_confusion_matrix(
    pairs: List[Tuple[ObjectAnnotation, ObjectAnnotation, ScalarMetricValue]],
    include_subclasses: bool,
    iou: float,
) -> ConfusionMatrixMetricValue:
    """
    Computes the confusion matrix for a list of object annotation pairs.
    Performs greedy matching of pairs.

    Args:
        pairs : A list of object annotation pairs with an iou score.
            This is used to determine matching priority (or if objects are matching at all) since objects can only be matched once.
        iou : iou threshold to deterine if objects are matching
    Returns:
        confusion matrix as a list: [TP,FP,TN,FN]
    """
    pairs.sort(key=lambda triplet: triplet[2], reverse=True)
    prediction_ids = set()
    ground_truth_ids = set()
    matched_predictions = set()
    matched_ground_truths = set()

    for ground_truth, prediction, agreement in pairs:
        prediction_id = id(prediction)
        ground_truth_id = id(ground_truth)
        prediction_ids.add(prediction_id)
        ground_truth_ids.add(ground_truth_id)

        if (
            agreement > iou
            and prediction_id not in matched_predictions
            and ground_truth_id not in matched_ground_truths
        ):
            if include_subclasses and (
                ground_truth.classifications or prediction.classifications
            ):
                if (
                    miou(
                        ground_truth.classifications,
                        prediction.classifications,
                        include_subclasses=False,
                    )
                    < 1.0
                ):
                    # Incorrect if the subclasses don't 100% agree then there is no match
                    continue
            matched_predictions.add(prediction_id)
            matched_ground_truths.add(ground_truth_id)
    tps = len(matched_ground_truths)

    fps = len(prediction_ids.difference(matched_predictions))
    fns = len(ground_truth_ids.difference(matched_ground_truths))
    # Not defined for object detection.
    tns = 0
    return [tps, fps, tns, fns]


def radio_confusion_matrix(
    ground_truth: Radio, prediction: Radio
) -> ConfusionMatrixMetricValue:
    """
    Calculates confusion between ground truth and predicted radio values

    Calculation:
        - TNs aren't defined because we don't know how many other classes exist
        - When P == L, then we get [1,0,0,0]
        - when P != L, we get [0,1,0,1]

    This is because we are aggregating the stats for the entire radio. Not for each class.
    Since we are not tracking TNs (P == L) only adds to TP.
    We are not tracking TNs because the number of TNs is equal to the number of classes which we do not know
    from just looking at the predictions and labels. Also TNs are necessary for precision/recall/f1.
    """
    key = get_identifying_key([prediction.answer], [ground_truth.answer])
    prediction_id = getattr(prediction.answer, key)
    ground_truth_id = getattr(ground_truth.answer, key)
    return [1, 0, 0, 0] if prediction_id == ground_truth_id else [0, 1, 0, 1]


def checklist_confusion_matrix(
    ground_truth: Checklist, prediction: Checklist
) -> ConfusionMatrixMetricValue:
    """
    Calculates agreement between ground truth and predicted checklist items:

    Calculation:
        - When a prediction matches a label that counts as a true postivie.
        - When a prediction was made and does not have a corresponding label this is counted as a false postivie
        - When a label does not have a corresponding prediction this is counted as a false negative

    We are also not tracking TNs since we don't know the number of possible classes
     (and they aren't necessary for precision/recall/f1).

    """
    key = get_identifying_key(prediction.answer, ground_truth.answer)
    schema_ids_pred = {getattr(answer, key) for answer in prediction.answer}
    schema_ids_label = {getattr(answer, key) for answer in ground_truth.answer}
    agree = schema_ids_label & schema_ids_pred
    all_selected = schema_ids_label | schema_ids_pred
    disagree = all_selected.difference(agree)
    fps = len({x for x in disagree if x in schema_ids_pred})
    fns = len({x for x in disagree if x in schema_ids_label})
    tps = len(agree)
    return [tps, fps, 0, fns]


def mask_confusion_matrix(
    ground_truths: List[ObjectAnnotation],
    predictions: List[ObjectAnnotation],
    include_subclasses: bool,
    iou: float,
) -> Optional[ScalarMetricValue]:
    """
    Computes confusion matrix metric for two masks

    Important:
        - If including subclasses in the calculation, then the metrics are computed the same as if it were object detection.
        - Each mask is its own instance. Otherwise this metric is computed as pixel level annotations.

    Args:
        ground_truths: List of ground truth mask annotations
        predictions: List of prediction mask annotations
    Returns:
        confusion matrix as a list: [TP,FP,TN,FN]
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return [0, len(predictions), 0, len(ground_truths)]
    elif has_no_annotations(ground_truths, predictions):
        return None

    pairs = _get_mask_pairs(ground_truths, predictions)
    return object_pair_confusion_matrix(
        pairs, include_subclasses=include_subclasses, iou=iou
    )


def ner_confusion_matrix(
    ground_truths: List[ObjectAnnotation],
    predictions: List[ObjectAnnotation],
    include_subclasses: bool,
    iou: float,
) -> Optional[ConfusionMatrixMetricValue]:
    """Computes confusion matrix metric between two lists of TextEntity objects

    Args:
        ground_truths: List of ground truth mask annotations
        predictions: List of prediction mask annotations
    Returns:
        confusion matrix as a list: [TP,FP,TN,FN]
    """
    if has_no_matching_annotations(ground_truths, predictions):
        return [0, len(predictions), 0, len(ground_truths)]
    elif has_no_annotations(ground_truths, predictions):
        return None
    pairs = _get_ner_pairs(ground_truths, predictions)
    return object_pair_confusion_matrix(pairs, include_subclasses, iou)



import json

from labelbox.data.annotation_types.data.generic_data_row_data import (
    GenericDataRowData,
)
import pytest

from labelbox.data.serialization import NDJsonConverter
from labelbox.types import (
    Label,
    MessageEvaluationTaskAnnotation,
    MessageSingleSelectionTask,
    MessageMultiSelectionTask,
    MessageInfo,
    OrderedMessageInfo,
    MessageRankingTask,
)


def test_message_task_annotation_serialization():
    with open("tests/data/assets/ndjson/mmc_import.json", "r") as file:
        data = json.load(file)

    labels = [
        Label(
            data=GenericDataRowData(
                uid="cnjencjencjfencvj",
            ),
            annotations=[
                MessageEvaluationTaskAnnotation(
                    name="single-selection",
                    extra={"uuid": "c1be3a57-597e-48cb-8d8d-a852665f9e72"},
                    value=MessageSingleSelectionTask(
                        message_id="clxfzocbm00083b6v8vczsept",
                        model_config_name="GPT 5",
                        parent_message_id="clxfznjb800073b6v43ppx9ca",
                    ),
                )
            ],
        ),
        Label(
            data=GenericDataRowData(
                uid="cfcerfvergerfefj",
            ),
            annotations=[
                MessageEvaluationTaskAnnotation(
                    name="multi-selection",
                    extra={"uuid": "gferf3a57-597e-48cb-8d8d-a8526fefe72"},
                    value=MessageMultiSelectionTask(
                        parent_message_id="clxfznjb800073b6v43ppx9ca",
                        selected_messages=[
                            MessageInfo(
                                message_id="clxfzocbm00083b6v8vczsept",
                                model_config_name="GPT 5",
                            )
                        ],
                    ),
                )
            ],
        ),
        Label(
            data=GenericDataRowData(
                uid="cwefgtrgrthveferfferffr",
            ),
            annotations=[
                MessageEvaluationTaskAnnotation(
                    name="ranking",
                    extra={"uuid": "hybe3a57-5gt7e-48tgrb-8d8d-a852dswqde72"},
                    value=MessageRankingTask(
                        parent_message_id="clxfznjb800073b6v43ppx9ca",
                        ranked_messages=[
                            OrderedMessageInfo(
                                message_id="clxfzocbm00083b6v8vczsept",
                                model_config_name="GPT 4 with temperature 0.7",
                                order=1,
                            ),
                            OrderedMessageInfo(
                                message_id="clxfzocbm00093b6vx4ndisub",
                                model_config_name="GPT 5",
                                order=2,
                            ),
                        ],
                    ),
                )
            ],
        ),
    ]

    res = list(NDJsonConverter.serialize(labels))

    assert res == data


def test_mesage_ranking_task_wrong_order_serialization():
    with pytest.raises(ValueError):
        (
            Label(
                data=GenericDataRowData(
                    uid="cwefgtrgrthveferfferffr",
                ),
                annotations=[
                    MessageEvaluationTaskAnnotation(
                        name="ranking",
                        extra={
                            "uuid": "hybe3a57-5gt7e-48tgrb-8d8d-a852dswqde72"
                        },
                        value=MessageRankingTask(
                            parent_message_id="clxfznjb800073b6v43ppx9ca",
                            ranked_messages=[
                                OrderedMessageInfo(
                                    message_id="clxfzocbm00093b6vx4ndisub",
                                    model_config_name="GPT 5",
                                    order=1,
                                ),
                                OrderedMessageInfo(
                                    message_id="clxfzocbm00083b6v8vczsept",
                                    model_config_name="GPT 4 with temperature 0.7",
                                    order=1,
                                ),
                            ],
                        ),
                    )
                ],
            ),
        )
